"""Utility functions for generating hashes and for parameter sets.

The hash of a parameter set is crucial to curifactory, as this hash gets
prefixed to filenames stored in the cache and so is used to determine whether a
particular artifact has already been computed for the given parameter set.

The basic idea of this hash computation is that some representation for every
parameter in a parameter set is retrieved, which is then turned into a string,
and the md5 hash of that string is then computed. The integer value of the
resulting md5 hashes of each parameter is added up, and the final integer is
turned back into a string hex "hash".

An important concept is the ability to modify any given parameter's
representation that is used for the md5 hash, and whether it's included as part
of the overall hash at all. Some types of objects in python by default will only
return a memory pointer when ``repr`` is called (which is the default mechanism
we use for getting a string representation,) which means that every time an
experiment is run, even if the parameters _should_ be the exact same, the hash
will be different. By setting a dictionary of ``hash_representations`` on the
parameter class, we can indivdiually control the representation computation for
each parameter. We can also set the parameter representation to ``None``, which
means it will be ignored for the purposes of the hash. This is useful for
"operational parameters", or configuration of an experimeriment that wouldn't
actually modify the artifacts.  (e.g. the number of gpu's to train an ML model
on and so forth.)

TODO: examples? (prob put this in the non-python-file docs)
"""

import hashlib
import json
import os
from copy import deepcopy
from dataclasses import field, fields, is_dataclass
from typing import Any, Callable, Union

PARAMETERS_BLACKLIST = ["name", "hash", "overwrite", "hash_representations"]
"""The default parameters on the ExperimentParameters class that we always
ignore as part of the hash."""


def set_hash_functions(*args, **kwargs):
    """Convenience function for easily setting the hash_representations dictionary
    with the appropriate dataclass field. Parameters passed to this function should
    be the same as the parameter name in the parameters class itself.

    You can either call this function and pass in a dictionary with the hashing functions,
    or pass each hashing function as a kwarg. If you pass in both a dictionary as the first
    positional arg and specify kwargs, the kwarg hashing functions will be added to the
    dictionary.

    Example:
        .. code-block:: python

            from dataclasses import dataclass
            from curifactory import ExperimentParameters
            from curifactory.params import set_hash_functions

            @dataclass
            class Params(ExperimentParameters):
                a: int = 0
                b: int = 0

                hash_representations: dict = set_hash_functions(
                    a = lambda self, obj: str(a)
                    b = None  # this means that b will _not be included in the hash_.
                )
    """
    if len(args) > 0:
        if type(args[0]) != dict:
            raise ValueError(
                "If providing a positional arg to set_hash_functions, it must be a dictionary."
            )
        return field(default_factory=lambda: {**(args[0]), **kwargs}, repr=False)
    return field(default_factory=lambda: dict(**kwargs), repr=False)


def get_parameter_hash_value(param_set, param_name: str) -> tuple[str, Any]:
    """Determines which hashing representation mechanism to use for the specified
    parameter, computes the result of the mechanism, and returns both.

    This function takes any overriding ``hash_representations`` into account. The list of mechanisms
    it attempts to use to get a hashable representation of the parameter in order are:

    1. Skip any blacklisted internal curifactory parameters that shouldn't affect the hash.
        This includes `name`, `hash`, `overwrite`, and the `hash_representations` attribute
        itself.
    2. If the value of the parameter is ``None``, skip it. This allows default-ignoring
        new parameters.
    3. If there's an associated hashing function in ``hash_representations``, call that,
        passing in the entire parameter set and the current value of the parameter to
        be hashed
    4. If a parameter is another dataclass, recursively ``get_paramset_hash_values`` on it.
        Note that if this is unintended functionality, and you need the default
        dataclass ``repr`` for any reason, you can override it with the following:

        .. code-block:: python

            import curifactory as cf

            @dataclass
            class Params(cf.ExperimentParameters):
                some_other_dataclass: OtherDataclass = None

                hash_representations = cf.set_hash_functions(
                    some_other_dataclass = lambda self, obj: obj.__class__
                )
                ...

    5. If a parameter is a callable, by default it might turn up a pointer address
        (we found this occurs with torch modules), so use the ``__qualname__``
        instead.
    6. Otherwise just use the normal ``repr``.

    Args:
        parameter_set: The parameter set (dataclass instance) to get the requested parameter from.
        parameter_name (str): The name of the parameter to get the hashable representation of.

    Returns:
        A tuple where the first element is the strategy used to compute the hashable representation,
        and the second element is that computed representation.
    """
    value = getattr(param_set, param_name)

    # 1. skip things we apriori know we don't want included
    if param_name in PARAMETERS_BLACKLIST:
        return ("SKIPPED: blacklist", None)

    # 2. see if user has specified how to handle the hash representation
    if (
        hasattr(param_set, "hash_representations")
        and param_name in param_set.hash_representations
    ):
        if param_set.hash_representations[param_name] is None:
            return ("SKIPPED: set to None in hash_representations", None)
        return (
            f"param_set.hash_representations['{param_name}'](param_set, {param_name})",
            param_set.hash_representations[param_name](param_set, value),
        )

    # 3. if the value of the argument is none, ignore it. This is so that we can default
    # arguments to not be included without setting the hash function for it, and may allow
    # fancier mechanisms in the future to better allow reproducing old experiments using a
    # parameters class that has since been added to.
    elif value is None:
        return ("SKIPPED: value is None", None)

    # -- some sane default hashing mechanisms --

    # 4. if it's a dataclass, recursively call get_param_set_hash_values on it, this allows
    # user to separate out subsets of parameters and still set custom hashing functions
    # on those subsets if they want.
    elif is_dataclass(value):
        return (
            f"get_param_set_hash_values(param_set, {param_name})",
            get_param_set_hash_values(value),
        )

    # 5. use the function name if it's a callable, rather than a pointer address
    elif isinstance(value, Callable):
        return (f"{param_name}.__qualname__", value.__qualname__)

    # 6. otherwise just use the default representation!
    return (f"repr({param_name})", repr(value))


def get_param_set_hash_values(param_set) -> dict[str, tuple[str, Any]]:
    """Collect the hash representations from every parameter in the passed parameter set.

    This essentially just calls ``get_parameter_hash_value`` on every parameter.

    Returns:
        A dictionary keyed by the string parameter names, and the value the dry tuple result
        from ``get_parameter_hash_value``.
    """
    # TODO: raise_error if param_set is not a dataclass?
    return {
        param.name: get_parameter_hash_value(param_set, param.name)
        for param in fields(param_set)
    }


def _compute_hash_part(hash_representations: dict[str, tuple[str, Any]]) -> int:
    """Recursive computation for the integer value of the hash of a passed hash_values dictionary."""
    hash_total = 0
    for hash_key, (hash_rep, hash_rep_value) in hash_representations.items():
        if hash_rep_value is None:
            continue

        # make sure to recursively compute on any subdataclasses
        if hash_rep.startswith("get_param_set_hash_values"):
            hash_total += _compute_hash_part(hash_rep_value)
        else:
            # Note that we concatenate the string of the value with the hash key, otherwise if two parameters had eachother's
            # values in another parameter set, they'd compute the same hash which is decidedly not correct.
            hash_hex = hashlib.md5(f"{hash_key}{hash_rep_value}".encode()).hexdigest()
            hash_total += int(hash_hex, 16)
    return hash_total


def compute_hash(hash_representations: dict[str, tuple[str, Any]]) -> str:
    """Returns a combined order-independent md5 hash of the passed representations.

    We do this by individually computing a hash for each item, and add the integer values up,
    turning the final number into a hash string.  this ensures that the order in which
    things are hashed won't change the hash as long as the values themselves are
    the same.
    """

    hash_total = _compute_hash_part(hash_representations)
    final_hash = f"{hash_total:x}"  # convert to a hexadecimal hash string
    return final_hash


# TODO: (3/10/2023) allow flag to still at least show the values of ignored parameters
def param_set_string_hash_representations(param_set) -> dict[str, str]:
    """Get the hash representation of a parameter set into a json-dumpable dictionary.

    This is used both in the output report as well as in the params registry.
    """
    hash_reps = get_param_set_hash_values(param_set)
    rep_dictionary = {}
    skipped = {}
    for key, rep_tuple in hash_reps.items():
        if key == "name":
            rep_dictionary[key] = param_set.name
        # check for a sub-dataclass, which might have ignored params of its own
        elif rep_tuple[0].startswith("get_param_set_hash_values"):
            rep_dictionary[key] = param_set_string_hash_representations(
                getattr(param_set, key)
            )
        elif rep_tuple[1] is not None:
            rep_dictionary[key] = str(rep_tuple[1])
        elif key in PARAMETERS_BLACKLIST:
            continue
        else:
            try:
                # separately handle a sub-dataclass, since we won't get the right strategy if it was skipped
                if is_dataclass(getattr(param_set, key)):
                    skipped[key] = param_set_string_hash_representations(
                        getattr(param_set, key)
                    )
                else:
                    skipped[key] = str(getattr(param_set, key))
            except:  # noqa: E722
                skipped[key] = None
    if len(skipped) > 0:
        rep_dictionary["IGNORED_PARAMS"] = skipped
    return rep_dictionary


def hash_param_set(
    param_set,
    store_in_registry: bool = False,
    registry_path: str = None,
    dry: bool = False,
) -> Union[str, dict]:
    """Returns a hex string representing the passed arguments, optionally recording
    the parameters and hash in the params registry.

    Note that this hash is computed once and then stored on the parameter set. If values
    on parameter set are changed and ``hash_param_set`` is called again, it won't be reflected
    in the hash.

    Args:
        param_set (ExperimentParameters): The argument set to hash.
        registry_path (str): The location to keep the :code:`params_registry.json`.
            If this is ``None``, ignore ``store_in_registry``.
        store_in_registry (bool): Whether to update the params registry with the passed
            arguments or not.
        dry (bool): If ``True``, don't store and instead return a dictionary
            with each value as the tuple that contains the strategy used to compute
            the values to be hashed as well as the output from that hashing function
            code. Useful for debugging custom hashing functions.

    Returns:
        The hash string computed from the arguments, or the dictionary of hashing functions
        if ``dry`` is ``True``. (The output from ``get_param_set_hash_values``)
    """
    hash_reps = get_param_set_hash_values(param_set)
    if dry:
        return hash_reps

    # if the hash has already been calculated/stored in the parameter set, use that
    # NOTE: I think this is also what allows someone to manually set a hash? This
    # functionality is untested
    if param_set.hash is not None:
        hash_str = param_set.hash
    else:
        # otherwise compute the integer for the hash representations and get the hex
        # string for it
        hash_str = compute_hash(hash_reps)

    # handle saving the hash and associated parameters in the registry
    if store_in_registry and registry_path is not None:
        registry_path = os.path.join(registry_path, "params_registry.json")
        registry = {}

        if os.path.exists(registry_path):
            with open(registry_path) as infile:
                registry = json.load(infile)

        reps_dictionary = param_set_string_hash_representations(param_set)
        reps_dictionary["_DRY_REPS"] = hash_reps
        registry[hash_str] = reps_dictionary
        with open(registry_path, "w") as outfile:
            json.dump(registry, outfile, indent=4, default=lambda x: str(x))

    return hash_str


# TODO: (3/5/2023) do I need the same default registry_path to None logic here as in hash_param_set?
def add_params_combo_hash(
    active_record, records_list, registry_path: str, store_in_registry: bool = False
):
    """Returns a hex string representing the the combined parameter set hashes from the
    passed records list. This is mainly used for getting a hash for an aggregate stage,
    which may not have a meaningful argument set of its own.

    Args:
        active_record (Record): The currently in-use record (likely owned by the aggregate
            stage.)
        records_list (List[Record]): The list of records to include as part of the resulting
            hash.
        registry_path (str): The location to keep the :code:`params_registry.json`.
        store_in_registry (bool): Whether to update the params registry with the passed
            records or not.

    Returns:
        The hash string computed from the combined record arguments.
    """

    hashes = []
    for agg_record in records_list:
        if agg_record.params is not None:
            hashes.append(agg_record.params.hash)
        else:
            hashes.append("None")
    hashes = sorted(hashes)

    hashes_for_key = deepcopy(hashes)
    active_key = "None"
    if active_record.params is not None:
        active_key = active_record.params.hash
    hashes_for_key.insert(0, active_key)

    hash_key = str(hashes_for_key)
    hash_str = hashlib.md5(hash_key.encode()).hexdigest()

    if store_in_registry:
        registry_path = os.path.join(registry_path, "params_registry.json")
        registry = {}

        if os.path.exists(registry_path):
            with open(registry_path) as infile:
                registry = json.load(infile)

        registry[hash_str] = {"active": active_key, "arg_list": hashes}
        with open(registry_path, "w") as outfile:
            json.dump(registry, outfile, indent=4, default=lambda x: str(x))
    return hash_str



from pytest_mock import mocker  # noqa: F401 -- flake8 doesn't see it's used as fixture

import curifactory.utils
from curifactory import docker
from curifactory.experiment import run_experiment


def test_build_docker(mocker):  # noqa: F811 -- mocker has to be passed in as fixture
    """This is mostly just to make sure the docker command isn't changed from what we
    expect without a notification."""

    mocker.patch("curifactory.utils.run_command")
    docker.build_docker("test_name", "./test_cache_folder", "some_version_string")

    curifactory.utils.run_command.assert_called_once_with(
        [
            "docker",
            "build",
            "-f",
            "docker/dockerfile",
            "--tag",
            "test_name:some_version_string",
            "--tag",
            "test_name",
            "--build-arg",
            "run_folder=./test_cache_folder",
            ".",
            "--progress",
            "tty",
        ]
    )


def test_build_docker_gets_correct_run_folder(
    mocker,  # noqa: F811 -- mocker has to be passed in as fixture
    configured_test_manager,
):
    """The name of the run folder passed to build_docker should be correct and not have the last
    letter cut off."""
    # mocker.patch("curifactory.utils.run_command")
    mocker.patch("curifactory.utils.run_command")

    results, manager = run_experiment(
        "basic", ["params1"], mngr=configured_test_manager, build_docker=True
    )
    run_num = manager.experiment_run_number
    datestr = manager.run_timestamp.strftime("%Y-%m-%d")
    curifactory.utils.run_command.assert_called_once_with(
        [
            "docker",
            "build",
            "-f",
            "docker/dockerfile",
            "--tag",
            f"basic:{run_num}_{datestr}",
            "--tag",
            "basic",
            "--build-arg",
            f"run_folder=test/examples/data/runs/test_{run_num}_{manager.run_timestamp.strftime(curifactory.utils.TIMESTAMP_FORMAT)}",
            ".",
            "--progress",
            "tty",
        ]
    )



# flake8: noqa

# make all submodules directly accessible from a single curifactory import
from curifactory import (
    caching,
    docker,
    experiment,
    hashing,
    manager,
    params,
    procedure,
    record,
    reporting,
    staging,
    store,
    utils,
)

# make super important things accessible directly off of the top level module
from curifactory.args import ExperimentArgs  # TODO: remove once fully deprecated
from curifactory.caching import Lazy
from curifactory.hashing import set_hash_functions
from curifactory.manager import ArtifactManager
from curifactory.params import ExperimentParameters
from curifactory.procedure import Procedure
from curifactory.record import Record
from curifactory.staging import (
    CachersMismatchError,
    EmptyCachersError,
    InputSignatureError,
    OutputSignatureError,
    aggregate,
    stage,
)

__version__ = "0.15.1"



import os
from dataclasses import dataclass
from enum import IntEnum

from graphviz import Digraph

import curifactory as cf
from curifactory import reporting
from curifactory.caching import JsonCacher, PickleCacher
from curifactory.reporting import JsonReporter, _add_record_subgraph, render_reportable


def test_reportables_cached(configured_test_manager):
    """Re-running a stage with cached outputs and reportables should still reload the old reportables."""

    @cf.stage(None, ["stuff"], [JsonCacher])
    def stage_with_reportables(record):
        dictionary = {"thing1": "testing"}
        record.report(JsonReporter(dictionary))

        with open(os.path.join(record.manager.cache_path, "stage_ran"), "w") as outfile:
            outfile.write("\n")

        return dictionary

    ran_path = os.path.join(configured_test_manager.cache_path, "stage_ran")

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    stage_with_reportables(r0)

    assert len(configured_test_manager.reportables) == 1
    assert os.path.exists(ran_path)
    os.remove(ran_path)

    r1 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    stage_with_reportables(r1)

    assert len(configured_test_manager.reportables) == 2
    assert not os.path.exists(ran_path)


def test_no_angle_brackets_in_report_argset_dump(configured_test_manager):
    """The output pre tag in the report argset dump should not contain un-escaped angle brackets."""

    class MyEnum(IntEnum):
        thing1 = 0
        thing2 = 2

    @dataclass
    class MyArgs(cf.ExperimentParameters):
        thing: MyEnum = MyEnum.thing1

    cf.Record(configured_test_manager, MyArgs())
    lines = reporting.render_report_argset_dump(configured_test_manager)
    all_text = "".join(lines[2:-1])
    assert "<" not in all_text
    assert ">" not in all_text


def test_reportable_render_uses_qualified_name_in_title(configured_test_manager):
    """Reportable rendering should use the fully qualified name in the link and the title"""

    @cf.stage(None, ["test_output"], [PickleCacher])
    def basic_reportable(record):
        record.report(JsonReporter({"test": "hello world"}))
        return "test"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    basic_reportable(r0)

    reportable = configured_test_manager.reportables[0]
    html = render_reportable(
        reportable, configured_test_manager, configured_test_manager.cache_path
    )
    assert reportable.qualified_name in html[1]  # link
    assert reportable.qualified_name in html[2]  # title


def test_detailed_record_subgraph_aggregate_doesnot_include_previous_artifacts(
    configured_test_manager,
):
    """A detailed subgraph map for an aggregate shouldn't show the nameless artifact
    nodes that are coming from other records."""

    @cf.stage(None, ["test_output"])
    def test(record):
        return "test"

    @cf.aggregate(inputs=["test_output"], outputs=["final"])
    def agg(record, records, test_output):
        return "things"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test1"))
    r1 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test2"))
    r2 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test3"))

    test(r0)
    test(r1)
    agg(r2, [r0, r1])

    dot = Digraph()
    dot.attr(compound="true")
    dot.attr(nodesep=".2")
    dot.attr(ranksep=".2")

    _add_record_subgraph(dot, 2, r2, configured_test_manager)

    # count connections
    conn_count = 0
    for line in dot.body:
        if "->" in line:
            conn_count += 1
    assert conn_count == 1



"""Classes for various caching strategies, known as "cachers"

This is handled through a base :code:`Cacheable` class, and each "strategy"
cacher class extends it.

Note that there are effectively two ways to use cacheables -
# TODO finish this
"""

import copy
import json
import logging
import os
import pickle
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Literal, Optional, Union

import pandas as pd

from curifactory import hashing, utils


class Lazy:
    """A class to indicate a stage output as a lazy-cache object - curifactory will
    attempt to keep this out of memory as much as possible, immediately caching and deleting,
    and loading back into memeory only when needed.

    This object is used by "wrapping" a stage output string with the class.

    Example:
        .. code-block:: python

            @stage(inputs=None, outputs=["small_output", Lazy("large_output")], cachers=[PickleCacher]*2)
            def some_stage(record: Record):
                ...

    Args:
        name (str): the name of the output to put into state.
        resolve (bool): Whether this lazy object should automatically reload the initial
            object when accessed from state. By default this is ``True`` - when a stage specifies
            the string name as an input and this object is requested from the record state, it
            loads and passes in the originally stored object. If set to ``False``, the stage
            input will instead be populated with the lazy object itself, giving the inner stage
            code direct access to the cacher. This is useful if you need to keep objects out of
            memory and just want to refer to the cacher path (e.g. to send this path along
            to an external CLI/script.)
    """

    def __init__(self, name: str, resolve: bool = True):
        self.name = name
        self.cacher: Cacheable = None
        self.resolve = resolve

    def load(self):
        return self.cacher.load()

    def __str__(self):
        return self.name


# NOTE: this isn't a requirement for DAGs to work, leaving here as a reminder, but deciding if
# this versus simply a pathcacher works better should be a different issue
# class Ref:
#     """A 'reference' output is very similar to the ``Lazy`` class, but handled differently semantically by
#     curifactory. A ``Ref`` is assumed to be a load-only output, where the stage that creates it is assumed
#     to handle creating and saving the object, and the cacher associated with the ``Ref`` is only in charge
#     of loading it when and if it's needed by later stages.

#     This is useful in cases where a stage is making an external system call to a script that's creating some
#     file, and we only want to directly bring the results of that file into curifactory rather than loading
#     and returning it in the stage, for it to just be re-saved again by the cacher.


#     Example:
#         .. code-block:: python

#             @stage(outputs=[Ref("some_json")], cachers=[JsonCacher])
#             def save_outside_of_cacher(record):
#                 pass


#                 # TODO

#     """

#     def __init__(self, name: str, resolve: bool = True):
#         self.name = name
#         self.cacher: Cacheable = None
#         self.resolve = resolve

#     def load(self):
#         return self.cacher.load()

#     def __str__(self):
#         return self.name


class Cacheable:
    """The base caching class, any caching strategy should extend this.

    Args:
        path_override (str): Use a specific path for the cacheable, rather than
            automatically setting it based on name etc.
        name (str): The obj name to use in automatically constructing a path. If
            a cacheable is used in stage header, this is automatically provided as
            the output string name from the stage outputs list.
        subdir (str): An optional string of one or more nested subdirectories to
            prepend to the artifact filepath.  This can be used if you want to
            subdivide cache and run artifacts into logical subsets, e.g. similar to
            https://towardsdatascience.com/the-importance-of-layered-thinking-in-data-engineering-a09f685edc71.
        prefix (str): An optional alternative prefix to the experiment-wide
            prefix (either the experiment name or custom-specified experiment
            prefix). This can be used if you want a cached object to work easier
            across multiple experiments, rather than being experiment specific.
            WARNING: use with caution, cross-experiment caching can mess with
            provenance.
        extension (str): The filetype extension to add at the end of the path.
        record (Record): The current record this cacheable is caching under.
            This can be used to get a copy of the current args instance and is also
            how artifact metadata is collected.
        track (bool): whether to include returned path in a store full copy or not.

    Note:
        It is strongly recommended that any subclasses of Cacheable take ``**kwargs`` in init and pass
        along to ``super()``:

        .. code-block:: python

            class CustomCacher(cf.Cacheable):
                def __init__(self, path_override: str = None, custom_attribute: Any = None, **kwargs):
                    super().__init__(path_override, **kwargs)
                    self.some_custom_attribute = custom_attribute

        This allows consistent handling of paths in the parent ``get_path()`` and ``check()`` functions.

        If no custom attributes are needed, also pass in *args, so path_override can be specified without
        a kwarg:

            .. code-block:: python

                class CustomCacher(cf.Cacheable):
                    def __init__(self, *args, **kwargs):
                        super().__init__(*args, extension=".custom", **kwargs)

    """

    def __init__(
        self,
        path_override: str = None,
        name: str = None,
        subdir: str = None,
        prefix: str = None,
        extension: str = None,
        record=None,
        track: bool = True,
    ):
        self.path_override = path_override
        """Use a specific path for the cacheable, rather than automatically setting it based on name etc."""
        self.name = name
        """The obj name to use in automatically constructing a path. If a cacheable is used in stage header, this is automatically
        provided as the output string name from the stage outputs list."""
        self.subdir = subdir
        """An optional string of one or more nested subdirectories to prepend to the artifact filepath.
        This can be used if you want to subdivide cache and run artifacts into logical subsets, e.g. similar to
        https://towardsdatascience.com/the-importance-of-layered-thinking-in-data-engineering-a09f685edc71.  """
        self.prefix = prefix
        """An optional alternative prefix to the experiment-wide prefix (either the experiment name or
        custom-specified experiment prefix). This can be used if you want a cached object to work easier across
        multiple experiments, rather than being experiment specific. WARNING: use with caution, cross-experiment
        caching can mess with provenance.  """
        self.extension = extension
        """The filetype extension to add at the end of the path. (Optional, automatically used as suffix in get_path if provided)"""
        self.record = None
        """The current record this cacheable is caching under. This can be used to get a copy of the current args instance and is also
        how artifact metadata is collected."""
        self.track = track
        """Whether to store the artifact this cacher is used with in the run folder on store-full runs or not."""
        self.cache_paths: list[str] = []
        """The running list of paths this cacher is using, as appended by ``get_path``."""
        self.metadata: dict = None
        """Metadata about the artifact cached with this cacheable."""
        self.extra_metadata: dict = {}
        """``collect_metadata`` uses but does not overwrite this, placing into the `extra` key
        in the actual metadata. This can be used by the cacher's save function to store additional
        information that would then be available if the 'load' function calls ``load_metadata()``."""
        self.stage: str = None
        """The stage associated with this cacher, if applicable."""

        if record is not None:
            self.set_record(record)

    def _resolve_suffix(
        self, path: str, suffix: str, add_extension: bool = True
    ) -> str:
        """Handles determining the final suffix of a path based on the extension of
        this cacheable and whether a specific suffix was requested or not."""

        # this is the logic to add an extension if no suffix supplied and the
        # object name/override path doesn't already contain the extension
        if suffix is None and (
            self.extension is not None
            and self.extension != ""
            and not path.endswith(self.extension)
            and add_extension
        ):
            suffix = self.extension
        elif suffix is None:
            suffix = ""

        return suffix

    def get_path(self, suffix=None) -> str:
        """Retrieve the full filepath to use for saving and loading. This should be called in the ``save()`` and
        ``load()`` implementations.

        Args:
            suffix (str): The suffix to append to the path. If not set, this will default to the cachable's extension.

        Note:
            If ``path_override`` is set on this cacher, this cacher _does not handle storing in the full store
            directory._ The assumption is that either you're referring to a static external path (which doesn't make
            sense to copy), or you're manually passing in a ``record.get_path`` in which case the record has already
            dealt with any logic necessary to add the path to the record's ``unstored_tracked paths`` which get copied
            over. Note also that this can be problematic for cachers that store multiple files since anything that isn't
            the path_override won't get copied. **For multiple file cachers you should use ``name``/``subdir``/``prefix``
            instead of setting a ``path_override``.**
        """
        path = None
        if self.path_override is not None:
            path = self.path_override

            # if the path_override has the extension in it already, remove it to handle
            # suffix addition consistently with non-path_override case
            # (the extension will be re-added in as necessary)
            # NOTE: we only do this if suffix is specified, otherwise we assume
            # path override is exactly what's desired. This is important because if
            # someone calls record.get_path and passes it in here, that get_path won't
            # auto handle an extension, and we need this get_path to return the same
            # thing as expected from record.get_path for tracking to work.
            if (
                self.extension != ""
                and self.extension is not None
                and path.endswith(self.extension)
                and suffix is not None
            ):
                path = path[: path.rindex(self.extension)]

            suffix = self._resolve_suffix(path, suffix, False)
            path = path + suffix
        else:
            if self.record is None:
                raise RuntimeError(
                    "Trying to call get_path on a cacher with no record and no path_override. Either pass a path directly into this cacher, or set cacher.record = some_record"
                )
            if self.name is None:
                raise RuntimeError(
                    "Trying to call get_path on a cacher with no object name and no path_overide. Either pass a path directly into this cacher, or set cacher.name = 'objectname'"
                )

            suffix = self._resolve_suffix(self.name, suffix)
            obj_name = self.name + suffix
            path = self.record.get_path(
                obj_name,
                subdir=self.subdir,
                prefix=self.prefix,
                stage_name=self.stage,
                track=self.track,
            )
        if path not in self.cache_paths:
            self.cache_paths.append(path)
        return path

    def get_dir(self, suffix=None) -> str:
        """Returns a path for a directory with the given suffix (if provided), appropriate for use in a ``save``
        and ``load`` function. This will create any subdirectories in the path if they don't exist.
        """
        path = None
        if suffix is None:
            suffix = ""

        if self.path_override is not None:
            path = self.path_override + suffix
        else:
            path = self.record.get_dir(
                self.name + suffix,
                subdir=self.subdir,
                prefix=self.prefix,
                stage_name=self.stage,
                track=self.track,
            )
        os.makedirs(path, exist_ok=True)
        return path

    def set_record(self, record):
        self.record = record
        if not self.record.manager.map_mode:
            self.collect_metadata()

    def collect_metadata(self):
        if self.record is None:
            raise RuntimeError(
                "Cannot collect metadata from a cacher with no associated record. Assign cacher.record if metadata is needed."
            )

        if self.record.manager.run_info is not None:
            manager_run_info = copy.copy(self.record.manager.run_info)
        else:
            logging.warning(
                "Manager does not currently have run info, artifact metadata will not contain specific experiment run data."
            )
            manager_run_info = {}
        if "status" in manager_run_info:
            del manager_run_info[
                "status"
            ]  # it's always going to be incomplete at this point

        # input_records are record references, need to resolve similarly to maps in report
        input_record_names = [
            record.get_reference_name() for record in self.record.input_records
        ]

        self.metadata = dict(
            artifact_generated=datetime.now().strftime(utils.TIMESTAMP_FORMAT),
            params_hash=self.record.get_hash(),
            params_name=self.record.params.name
            if self.record.params is not None
            else None,
            record_name=self.record.get_reference_name(),
            stage=self.record.manager.current_stage_name,
            artifact_name=self.name,
            cacher_type=str(type(self)),
            record_prior_stages=self.record.stages[:-1],
            prior_records=input_record_names,
            params=hashing.param_set_string_hash_representations(self.record.params)
            if self.record.params is not None
            else None,
            extra=self.extra_metadata,  # cachers can store any additional info here they want.
            manager_run_info=manager_run_info,
        )

    def save_metadata(self):
        metadata_path = self.get_path("_metadata.json")
        with open(metadata_path, "w") as outfile:
            if self.metadata is None:
                # this either means we haven't collected metadata, or this is save() being called inline
                # logging.warning(
                #     "Cacher metadata hasn't been collected or has no associated record. Only saving extra_metadata fields."
                # )
                json.dump(
                    dict(extra=self.extra_metadata), outfile, indent=2, default=str
                )
            else:
                self.metadata["extra"].update(self.extra_metadata)
                json.dump(self.metadata, outfile, indent=2, default=str)

    def load_metadata(self) -> dict:
        metadata_path = self.get_path("_metadata.json")
        if os.path.exists(metadata_path):
            with open(metadata_path) as infile:
                self.metadata = json.load(infile)
                self.extra_metadata = self.metadata["extra"]
        return self.metadata

    def check(self) -> bool:
        """Check to see if this cacheable needs to be written or not.

        Note:
            This function will always return False if the args are ``None``.

        Returns:
            ``True`` if we find the cached file and the current parameter set
            doesn't specify to overwrite, otherwise ``False``.
        """
        logging.debug("Searching for cached file at '%s'...", self.get_path())
        if os.path.exists(self.get_path()):
            if (
                self.record is not None
                and self.record.params is None
                and self.record.is_aggregate
                and len(self.record.input_records) > 0
            ):
                logging.debug(
                    "Aggregate stage has no args, checking input records to determine overwrite"
                )

                # TODO: (05/16/2022) in principle we could have a function to do an
                # overwrite check on the record itself, that way it could be "recursive"
                # if we reach back through previous records that are aggregate as well

                # if it's an aggregate stage with no provided args, we check
                # to see if any of the associated records are set to overwrite
                for record in self.record.input_records:
                    if record.params is not None and record.params.overwrite:
                        logging.debug("Record with overwrite found")
                        return False
                # we made it through each record and they weren't overwrite, we're good
                logging.debug("No records had overwrite, will use cache")
                if self.record.manager.map_mode:
                    logging.debug("Cached object '%s' found", self.get_path())
                else:
                    logging.info("Cached object '%s' found", self.get_path())
                return True
            elif (
                self.record is not None
                and self.record.params is not None
                and not self.record.params.overwrite
            ):
                if self.record.manager.map_mode:
                    logging.debug("Cached object '%s' found", self.get_path())
                else:
                    logging.info("Cached object '%s' found", self.get_path())
                return True
            elif self.record is None:
                # if we don't have a record (e.g. running check on a manual cacher with a
                # path override specified), don't worry about overwrite logic.
                logging.info("Cached object '%s' found", self.get_path())
                return True
            else:
                logging.debug("Object found, but overwrite specified in args")
                return False
        logging.debug("Cached file not found")
        return False

    def load(self):
        """Load the cacheable from disk.

        Note:
            Any subclass is **required** to implement this.
        """
        raise NotImplementedError(
            "Cacheable class does not have a load function implemented"
        )

    def save(self, obj):
        """Save the passed object to disk.

        Note:
            Any subclass is **required** to implement this.
        """
        raise NotImplementedError(
            "Cacheable class does not have a save function implemented"
        )


class JsonCacher(Cacheable):
    """Dumps an object to indented JSON."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, extension=".json", **kwargs)

    def load(self):
        with open(self.get_path()) as infile:
            obj = json.load(infile)
        return obj

    def save(self, obj) -> str:
        path = self.get_path()
        with open(path, "w") as outfile:
            json.dump(obj, outfile, indent=4, default=lambda x: str(x))
        return path


class PickleCacher(Cacheable):
    """Dumps an object to a pickle file."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, extension=".pkl", **kwargs)

    def load(self):
        with open(self.get_path(), "rb") as infile:
            obj = pickle.load(infile)
        return obj

    def save(self, obj) -> str:
        path = self.get_path()
        with open(path, "wb") as outfile:
            pickle.dump(obj, outfile)
        return path


class _PandasIOType(Enum):
    csv = "csv"
    json = "json"
    parquet = "parquet"
    pickle = "pkl"
    orc = "orc"
    hdf5 = "h5"
    excel = "xlsx"
    xml = "xml"


class PandasCacher(Cacheable):
    """Saves a pandas dataframe to selectable IO format.

    Args:
        format (str): Selected pandas IO format. Choices are:
            ("csv", "json", "parquet", "pickle",
            "orc", "hdf5", "excel", "xml")
        to_args (Dict): Dictionary of arguments to use in the pandas
            ``to_*()`` call.
        read_args (Dict): Dictionary of arguments to use in the pandas
            ``read_*()`` call.
    """

    def __init__(
        self,
        path_override: Optional[str] = None,
        format: Literal[
            "csv", "json", "parquet", "pickle", "orc", "hdf5", "excel", "xml"
        ] = "pickle",
        to_args: Optional[dict] = None,
        read_args: Optional[dict] = None,
        **kwargs,
    ):
        self.format = _PandasIOType[format]
        self.read_args = read_args
        self.to_args = to_args

        # set up default args for specific types, if none are provided
        if self.to_args is None:
            if self.format == _PandasIOType.json:
                self.to_args = {"double_precision": 15}
            elif self.format == _PandasIOType.hdf5:
                self.to_args = {"key": "df"}
            elif self.format == _PandasIOType.xml:
                self.to_args = {"index": False}
            else:
                self.to_args = {}

        if self.read_args is None:
            if self.format in (_PandasIOType.csv, _PandasIOType.excel):
                self.read_args = {"index_col": 0}
            else:
                self.read_args = {}

        if self.format == _PandasIOType.json:
            logging.warning(
                "Using this cacher is inadvisable for floating point data, as precision"
                "will be lost, creating the potential for different results when using"
                "cached values with this cacher as opposed to the first non-cached run."
            )

        super().__init__(
            path_override=path_override, extension=f".{self.format.value}", **kwargs
        )

    # TODO: 7/31/2023 - might be worth considering adding a function to "populate from metadata",
    # which would load the metadata for the specified path, and then populate both format and
    # read_args. This could make it a little simpler to use a pandas cacher inline in a notebook
    # for reading.

    def load(self):
        # select the appropriate pandas read function based on format
        if self.format == _PandasIOType.csv:
            pandas_read = pd.read_csv
        elif self.format == _PandasIOType.json:
            pandas_read = pd.read_json
        elif self.format == _PandasIOType.parquet:
            pandas_read = pd.read_parquet
        elif self.format == _PandasIOType.pickle:
            pandas_read = pd.read_pickle
        elif self.format == _PandasIOType.orc:
            pandas_read = pd.read_orc
        elif self.format == _PandasIOType.hdf5:
            pandas_read = pd.read_hdf
        elif self.format == _PandasIOType.excel:
            pandas_read = pd.read_excel
        elif self.format == _PandasIOType.xml:
            pandas_read = pd.read_xml
        else:
            raise RuntimeError(f"Invalid Pandas IO Type ({self.format.name}) selected")

        # double check that there's no pandas version mismatch
        # at some point it may be worth warning only on major/minor version difference
        self.load_metadata()
        if "pandas_version" in self.extra_metadata:
            if self.extra_metadata["pandas_version"] != pd.__version__:
                logging.warning(
                    "Attempting to use pandas v%s to load a dataframe that was initially saved with pandas v%s"
                    % (pd.__version__, self.extra_metadata["pandas_version"])
                )

        return pandas_read(self.get_path(), **self.read_args)

    def save(self, obj: pd.DataFrame) -> str:
        # select appropriate pandas write function based on format
        if self.format == _PandasIOType.csv:
            pandas_to = obj.to_csv
        elif self.format == _PandasIOType.json:
            pandas_to = obj.to_json
        elif self.format == _PandasIOType.parquet:
            pandas_to = obj.to_parquet
        elif self.format == _PandasIOType.pickle:
            pandas_to = obj.to_pickle
        elif self.format == _PandasIOType.orc:
            pandas_to = obj.to_orc
        elif self.format == _PandasIOType.hdf5:
            pandas_to = obj.to_hdf
        elif self.format == _PandasIOType.excel:
            pandas_to = obj.to_excel
        elif self.format == _PandasIOType.xml:
            pandas_to = obj.to_xml
        else:
            raise RuntimeError(f"Invalid Pandas IO Type ({self.format.name}) selected")
        # record relevant cacher information to help track down any issues if they arise
        self.extra_metadata["pandas_version"] = pd.__version__
        self.extra_metadata["to_args"] = self.to_args
        self.extra_metadata["read_args"] = self.read_args
        self.extra_metadata["format"] = self.format.name
        self.save_metadata()

        # write the thing!
        path = self.get_path()
        pandas_to(path, **self.to_args)
        return path


class PandasJsonCacher(PandasCacher):
    """Saves a pandas dataframe to JSON.

    Warning:

        Using this cacher is inadvisable for floating point data, as precision
        will be lost, creating the potential for different results when using
        cached values with this cacher as opposed to the first non-cached run.

    Args:
        to_json_args (Dict): Dictionary of arguments to use in the pandas
            ``to_json()`` call.
        read_json_args (Dict): Dictionary of arguments to use in the pandas
            ``read_json()`` call.

    Note:
        This is equivalent to using ``PandasCacher(format='json')``
    """

    def __init__(
        self,
        path_override: str = None,
        to_json_args: dict = dict(double_precision=15),
        read_json_args: dict = dict(),
        **kwargs,
    ):
        super().__init__(
            path_override=path_override,
            format="json",
            to_args=to_json_args,
            read_args=read_json_args,
            **kwargs,
        )


class PandasCsvCacher(PandasCacher):
    """Saves a pandas dataframe to CSV.

    Args:
        to_csv_args (Dict): Dictionary of arguments to use in the pandas
            ``to_csv()`` call.
        read_csv_args (Dict): Dictionary of arguments to use in the pandas
            ``read_csv()`` call.

    Note:
        This is equivalent to using ``PandasCacher(format='csv')``
    """

    def __init__(
        self,
        path_override: str = None,
        to_csv_args: dict = dict(),
        read_csv_args: dict = dict(index_col=0),
        **kwargs,
    ):
        super().__init__(
            path_override=path_override,
            format="csv",
            to_args=to_csv_args,
            read_args=read_csv_args,
            **kwargs,
        )


class FileReferenceCacher(Cacheable):
    """Saves a file path or list of file paths generated from a stage as a json file.
    The ``check`` function will check existence of all file paths.

    This is useful for instances where there may be a large number of files stored or
    generated to disk, as it would be unwieldy to return them all (or infeasible to keep them
    in memory) directly from the stage. When this cacher is checked for pre-existing,
    it tries to load the json file storing the filenames, and then checks for the existence of
    each path in that json file. If all of them exist, it will short-circuit computation.

    Using this cacher does mean the user is in charge of loading/saving the file paths correctly,
    but in some cases that may be desirable.

    This can also be used for storing a reference to a single file outside the normal cache.

    When combined with the ``get_dir`` call on the record, you can create a cached directory of
    files similarly to a regular cacher and simply keep a reference to them as part of the actual
    cacher process.

    Example:

        .. code-block:: python

            @stage(inputs=None, outputs=["many_text_files"], cachers=[FileReferenceCacher])
            def output_text_files(record):
                file_path = record.get_dir("my_files")
                my_file_list = [os.path.join(file_path, f"my_file_{num}") for num in range(20)]

                for file in my_file_list:
                    with open(file, 'w') as outfile:
                        outfile.write("test")

                return my_file_list
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, extension=".json", **kwargs)

    def check(self) -> bool:
        # check the file list file exists
        if not super().check():
            return False

        # load the file list and check each file
        # NOTE: we don't need to re-check args overwrite because that
        # would already have applied in the super check
        with open(self.get_path()) as infile:
            files = json.load(infile)

        if type(files) == list:
            for file in files:
                logging.debug("Checking from file list: '%s'" % file)
                if not os.path.exists(file):
                    return False
        else:
            if not os.path.exists(files):
                return False

        return True

    def load(self) -> Union[list[str], str]:
        with open(self.get_path()) as infile:
            files = json.load(infile)
        return files

    def save(self, files: Union[list[str], str]) -> str:
        path = self.get_path()
        with open(path, "w") as outfile:
            json.dump(files, outfile, indent=4)
        return path


class RawJupyterNotebookCacher(Cacheable):
    """Take a list of code cells (where each cell is a list of strings containing python code)
    and turn it into a jupyter notebook. This is useful in situations where you want each
    experiment to have some form of automatically populated analysis that a reportable wouldn't
    sufficiently cover, e.g. an interactive set of widgets or dashboard.

    Example:

        .. code-block:: python

            @stage(inputs=["results_path"], outputs=["exploration_notebook"], cachers=[RawJupyterNotebookCacher])
            def make_exploration_notebook(record, results_path):
                def convert_path(path):
                    '''A function to translate paths to local folder path.'''
                    p = Path(path)
                    p = Path(*p.parts[2:])
                    return str(p)

                cells = [
                    [
                        "# imports",
                        "from curifactory.caching import JsonCacher",
                    ],
                    [
                        "# load things",
                        f"cacher = JsonCacher('./{convert_path(results_path)})",
                        "results = cacher.load()",
                        "results_metadata = cacher.metadata",
                    ],
                    [
                        "# analysis",
                        "...",
                    ],
                ]

                return cells
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, extension=".ipynb", **kwargs)

    def load(self):
        cells_path = self.get_path("_cells.json")
        return JsonCacher(cells_path).load()

    def save(self, obj: list[list[str]]):
        """This saves the raw cell strings to a _cells.json, and then uses
        ipynb-py-convert to change the output python script into a notebook
        format."""
        notebook_path = Path(self.get_path())
        cells_path = self.get_path("_cells.json")
        script_path = notebook_path.with_suffix(".py")

        cell_text = "\n\n# %%\n".join(["\n".join(inner_cell) for inner_cell in obj])

        with open(script_path, "w") as outfile:
            outfile.write(cell_text)
        utils.run_command(["ipynb-py-convert", script_path, notebook_path])
        os.remove(script_path)

        JsonCacher(cells_path).save(obj)



"""Contains relevant classes for records, objects that track a persistant state
through some set of stages for a given parameter set."""

import copy
import logging
import os
import shutil
from typing import Any

from curifactory import hashing, utils
from curifactory.caching import Lazy
from curifactory.reporting import Reportable
from curifactory.utils import _warn_deprecation


class CacheAwareDict(dict):
    """A normal dictionary that will return resolved versions of Lazy objects when accessed."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.resolve = True
        """A flag for enabling/disabling auto-resolve. If this is False, this works just
        like a normal dictionary. This is necessary in the input checking part of a normal stage, as
        we make state access before determining if stage execution is required or not."""

    def __getitem__(self, key):
        item = super().__getitem__(key)
        if type(item) == Lazy and self.resolve and item.resolve:
            logging.debug("Auto-resolving lazy object '%s'..." % key)
            return item.cacher.load()
        else:
            return item


# TODO: make printing this dump the state out as a string
# TODO: handle if args is None
class Record:
    """A single persistent state that's passed between stages in a single "experiment line".

    Args:
        manager (ArtifactManager): The artifact manager this record is associated with.
        param_set: The parameter set (subclass of ``ExperimentParameters``) to apply to
            any stages this record is run through.
        hide (bool): If ``True``, don't add this record to the artifact manager.
    """

    def __init__(self, manager, param_set, hide=False):
        # TODO: would be nice to be able to initialize a record with a pre-defined state dictionary
        self.manager = manager
        """The ``ArtifactManager`` associated with this record."""
        self.args = None
        """DEPRECATED."""
        self.params = param_set
        """The parameter set (subclass of ``ExperimentParameters``) to apply to any stages
        this record is passed through."""
        self.state = CacheAwareDict()
        """The dictionary of all variables created by stages this record is passed through. (AKA 'Artifacts')
        All ``inputs`` from stage decorators are pulled from this dictionary, and all
        ``outputs`` are stored here."""
        self.state_artifact_reps: dict[str, int] = {}
        """Dictionary mimicking state that keeps an index to the associated artifact representation in
         manager's artifact representation list."""
        self.output = None
        """The returned value from the last stage that was run with this record."""
        self.stages = []
        """The list of stage names that this record has run through so far."""
        self.stage_suppress_missing: list[bool] = []
        """The list of whether the stage with the associated index in ``self.stages``
        specified ``suppress_missing_inputs`` or not. This is primarly needed for the
        DAG mapping phase."""
        self.stage_kwargs_keys: list[list[str]] = []
        """The keys of any kwargs passed into each stage from the experiment, the index associated
        with ``self.stages``. We don't store the full dictionary because there could potentially be
        large values and the user may not expect a reference to stick around etc. We really only
        need this for checking correctness of an experiment setup during DAG mapping phase."""
        self.stage_inputs_names: list[list[str]] = []
        """The names of the stage inputs as specified in the decorator. This is primarily for
        helping checking correctness during DAG mapping phase."""
        self.stage_inputs: list[list[int]] = []
        """A list of lists per stage with the state inputs that stage requested.
        These are lists of indices into state_artifact_reps."""
        self.stage_outputs: list[list[int]] = []
        """A list of lists per stage with the state outputs that stage produced.
        These are lists of indices into state_artifact_reps."""
        self.input_records: list[Record] = []
        """A list of any records used as input to this one. This mostly only occurs when aggregate
        stages are run."""
        # NOTE: these are actual record references
        self.is_aggregate = False
        """If this record runs an aggregate stage, we flip this flag to true to know we need to use the
        combo hash rather than the individual args hash."""
        self.combo_hash = None
        """This gets set on records that run an aggregate stage. This is set from ``utils.add_params_combo_hash.``"""
        self.unstored_tracked_paths: list[dict[str, str]] = []
        """Paths obtained with get_path/get_dir that should be copied to a full
        store folder. The last executed stage should manage copying anything
        listed here and then clearing it. This is a list of dicts that would be
        passed to the artifact manager's ``get_artifact_path`` function:
        (obj_name, subdir, prefix, and path)
        """
        self.stored_paths: list[str] = []
        """A list of paths that have been copied into a full store folder. These are
        the source paths, not the destination paths."""
        self.stage_cachers: list = None
        """A list of the initialized cachers set for the current stage, if any. This is so that a stage
        can get access to output path information if it needs."""

        self.set_hash()
        if not hide:
            self.manager.records.append(self)

    def __getattribute__(self, __name: str) -> Any:
        if __name == "args":
            _warn_deprecation(
                "'Record.args' is deprecated and will likely be removed in 0.16.0. Please use 'Record.params' instead."
            )
            return self.params
        return object.__getattribute__(self, __name)

    def store_tracked_paths(self):
        """Copy all of the recent relevant files generated (likely from the recently executing
        stage) into a store-full run. This is run automatically at the end of a stage.
        """
        if self.manager.store_full:
            for path_info in self.unstored_tracked_paths:
                name = path_info["obj_name"]
                subdir = path_info["subdir"]
                prefix = path_info["prefix"]
                path = path_info["path"]

                # don't duplicate if we've already stored it (this might occur from multiple get_path
                # calls on a cacher)
                if path in self.stored_paths:
                    continue

                # paths can get added to the list that don't exist from a cacher's check() call
                # (e.g. checking if reportables exist but a stage output no reportables.)
                # at least for now we silently skip these.
                if not os.path.exists(path):
                    continue

                # if the auto naming strategy isn't being used (only a filepath given, no name),
                # then use the last part of the filename (from the last '/') as the obj name
                if name is None:
                    name = os.path.basename(path)
                    # TODO: (3/22/2023) unclear if I need to set subdir and prefix to none or not

                store_path = self.manager.get_artifact_path(
                    name, record=self, subdir=subdir, prefix=prefix, store=True
                )
                logging.debug(f"Copying tracked path '{path}' to '{store_path}'...")
                if os.path.isdir(path):
                    shutil.copytree(path, store_path)
                else:
                    shutil.copy(path, store_path)

                # remember that we copied this path
                self.stored_paths.append(path)
        # I'm clearing unstored regardless of store full or not, because we may
        # eventually want to support something like --store-artifact, where we
        # selectively add specific things to the tracked paths, so we want tracked
        # paths to not build up things that are never going to be stored.
        self.unstored_tracked_paths = []

    def set_hash(self):
        """Establish the hash for the current parameter set (and set it on the parameter set instance)."""
        # NOTE: we used to set this directly in manager's get_path, but there's potentially weird effects and it's an
        # odd place to establish a hash that more correctly indicates a record than the parameter set themselves (e.g. like with
        # aggregate combo hashes)
        if self.params is not None and self.params.hash is None:
            self.params.hash = hashing.hash_param_set(
                self.params,
                store_in_registry=not (self.manager.dry or self.manager.parallel_mode),
                registry_path=self.manager.manager_cache_path,
            )

            if self.manager.store_full:
                hashing.hash_param_set(
                    self.params,
                    store_in_registry=not (
                        self.manager.dry or self.manager.parallel_mode
                    ),
                    registry_path=self.manager.get_run_output_path(),
                )

    def get_hash(self) -> str:
        """Returns either the hash of the parameter set, or the combo hash if this record is an aggregate."""
        if self.is_aggregate:
            return self.combo_hash
        elif self.params is not None:
            return self.params.hash
        else:
            return "None"

    def set_aggregate(self, aggregate_records):
        """Mark this record as starting with an aggregate stage, meaning the hash of all cached outputs produced
        within this record need to reflect the combo hash of all records going into it.
        """
        self.is_aggregate = True
        self.combo_hash = hashing.add_params_combo_hash(
            self,
            aggregate_records,
            self.manager.manager_cache_path,
            not (self.manager.dry or self.manager.parallel_mode),
        )
        if self.manager.store_full:
            hashing.add_params_combo_hash(
                self,
                aggregate_records,
                self.manager.get_run_output_path(),
                not (self.manager.dry or self.manager.parallel_mode),
            )

    def report(self, reportable: Reportable):
        """Add a reportable associated with this record, this will get added to the experiment run
        output report.

        Args:
            reportable (Reportable): The reportable to render on the final experiment report.
        """
        reportable.record = self
        reportable.stage = self.stages[-1]

        qualified_name = ""
        if reportable.record.is_aggregate:
            qualified_name = "(Aggregate)_"
        if reportable.record.params is not None:
            qualified_name += f"{reportable.record.params.name}_"
        qualified_name += f"{reportable.stage}_"

        if reportable.name is None:
            qualified_name += str(len(self.manager.reportables))
        else:
            qualified_name += reportable.name
        reportable.qualified_name = qualified_name

        self.manager.reportables.append(reportable)

    def make_copy(self, param_set=None, add_to_manager=True):
        """Make a new record that has a deep-copied version of the current state.

        This is useful for a long running procedure that creates a common dataset for
        many other stages, so that it can be replicated across multiple parameter sets without
        having to recompute for each individual parameter set.

        Note that state is really the only thing transferred to the new record, the stage and
        inputs/outputs lists will be empty.

        Also note that the current record will be added to the ``input_records`` of the new
        record, since it may draw on data in its state.

        Args:
            param_set: The new ``ExperimentParameters`` to apply to the new record. Leave as ``None``
                to retain the same parameter set as the current record.
            add_to_manager: Whether to automatically add this record to the current manager or not.

        Returns:
            A new record with the same state as this one, but under a different parameter set.
        """
        if param_set is None:
            param_set = self.params
        new_record = Record(self.manager, param_set, hide=(not add_to_manager))
        new_record.input_records = [self]
        new_record.state = copy.deepcopy(self.state)
        # we shallow copy because this new list needs to be modified with new artifact_reps without
        # modifying the one in the old record
        new_record.state_artifact_reps = copy.copy(self.state_artifact_reps)
        return new_record

    # TODO: should also take an optional 'sub-path' and 'extension'
    def get_path(
        self,
        obj_name: str,
        subdir: str = None,
        prefix: str = None,
        stage_name: str = None,
        track: bool = True,
    ) -> str:
        """Return a cache path with passed object name and the correct hash based on the parameter set.

        This should be equivalent to what a cacher for a stage should get. Note that this
        is calling the manager's ``get_path``, which will include the stage name. If calling
        this outside of a stage, it will include whatever stage was last run.

        Args:
            obj_name (str): the name to associate with the object as the last part of the filename.
            subdir (str): An optional string of one or more nested subdirectories to prepend to the artifact filepath.
                This can be used if you want to subdivide cache and run artifacts into logical subsets, e.g. similar to
                https://towardsdatascience.com/the-importance-of-layered-thinking-in-data-engineering-a09f685edc71.
            prefix (str): An optional alternative prefix to the experiment-wide prefix (either the experiment name or
                custom-specified experiment prefix). This can be used if you want a cached object to work easier across
                multiple experiments, rather than being experiment specific. WARNING: use with caution, cross-experiment
                caching can mess with provenance.
            stage_name (str): The associated stage for a path. If not provided, the currently
                executing stage name is used.
            track (bool): whether to include returned path in a store full copy
                or not. This will only work if the returned path is not altered
                by a stage before saving something to it.
        """
        path = self.manager.get_artifact_path(
            obj_name=obj_name,
            record=self,
            subdir=subdir,
            prefix=prefix,
            stage_name=stage_name,
        )
        if track:
            # TODO: (3/22/2023) do I need to also be storing stage name? These are always supposed
            # to be handled from the current stage only anyway, so the stage name should always
            # be the last one
            self.unstored_tracked_paths.append(
                dict(obj_name=obj_name, subdir=subdir, prefix=prefix, path=path)
            )
        return path

    def get_dir(
        self,
        dir_name_suffix: str,
        subdir: str = None,
        prefix: str = None,
        stage_name: str = None,
        track: bool = True,
    ) -> str:
        """Returns a cache path with the passed name and appropriate hash (similar to ``get_path``) and creates it as a directory.

        Args:
            dir_name_suffix (str): the name to add as a suffix to the created directory name.
            subdir (str): An optional string of one or more nested subdirectories to prepend to the artifact filepath.
                This can be used if you want to subdivide cache and run artifacts into logical subsets, e.g. similar to
                https://towardsdatascience.com/the-importance-of-layered-thinking-in-data-engineering-a09f685edc71.
            prefix (str): An optional alternative prefix to the experiment-wide prefix (either the experiment name or
                custom-specified experiment prefix). This can be used if you want a cached object to work easier across
                multiple experiments, rather than being experiment specific. WARNING: use with caution, cross-experiment
                caching can mess with provenance.
            stage_name (str): The associated stage for a path. If not provided, the currently
                executing stage name is used.
            track (bool): whether to include returned path in a store full copy
                or not. This will only work if the returned path is not altered
                by a stage before saving something to it.
        """
        dir_path = self.manager.get_artifact_path(
            obj_name=dir_name_suffix,
            record=self,
            subdir=subdir,
            prefix=prefix,
            stage_name=stage_name,
        )
        if track:
            self.unstored_tracked_paths.append(
                dict(
                    obj_name=dir_name_suffix,
                    subdir=subdir,
                    prefix=prefix,
                    path=dir_path,
                )
            )

        os.makedirs(dir_path, exist_ok=True)
        return dir_path

    def get_reference_name(self, map=False) -> str:
        """This returns a name describing the record, in the format 'Record [index on manager] (paramset name)

        This should be the same as what's shown in the stage map in the output report.
        """
        index = self.get_record_index(map)
        paramset_name = self.params.name if self.params is not None else "None"
        return f"Record {index} ({paramset_name})"

    def get_record_index(self, map=False) -> int:
        if map:
            record_list = self.manager.map.records
        else:
            record_list = self.manager.records
        for i, record in enumerate(record_list):
            if self == record:
                return i
        return -1


class MapArtifactRepresentation:
    def __init__(
        self,
        record_index: int,
        stage_name: str,
        name: str,
        cached: bool,
        metadata=None,
        cacher=None,
    ):
        # NOTE: we're not keeping track of record because when the map stuff
        # is transfered from the manager over into the DAG, we make a separate
        # record instance anyway.
        # NOTE: (4/5/2023) we may want to at least keep a record index.
        self.record_index = record_index
        self.stage_name = stage_name
        self.name = name
        self.cached = cached
        self.metadata = metadata
        self.cacher = cacher


class ArtifactRepresentation:
    """This is a shorthand string representation for an artifact stored in a record as well
    as output cache info. This is what gets displayed in the detailed experiment map in reports.

    This will try to include helpful information in the string representation, such as pandas/
    numpy shapes, or lengths where applicable.

    Args:
        record (Record): The record this artifact is stored in.
        name (str): The name of the artifact.
        artifact: The artifact itself.
    """

    def __init__(self, record, name, artifact, metadata=None, cacher=None):
        # TODO: (3/21/2023) possibly have "files" which would be cachers.cached_files?
        self.init_record = record
        self.name = name
        self.string = utils.preview_object(artifact)

        self.cacher = cacher
        self.metadata = metadata

        if artifact is None and metadata is not None and "preview" in metadata:
            self.string = metadata["preview"]

        self.file = "no file"

    def html_safe(self):
        """Removes special characters that'd break html, except it doesn't actually display
        correctly.

        (I think this is an issue with how graphviz renders outputs?)"""
        clean = (
            self.string.replace("<", "&lt;")
            .replace(">", "&gt;")
            .replace("&", "&amp;")
            .replace('"', "&quot;")
        )

        return clean



""" Helper and utility functions for the library. """

import json
import logging
import os
import platform
import shutil
import subprocess
import sys
import warnings

from rich import get_console, reconfigure
from rich.logging import RichHandler

TIMESTAMP_FORMAT = "%Y-%m-%d-T%H%M%S"
"""The datetime format string used for timestamps in experiment run reference names."""
CONFIGURATION_FILE = "curifactory_config.json"
"""The expected configuration filename."""

EDITORS = ["vim", "nvim", "emacs", "nano", "vi"]
"""The list of possible editor exec names to test for getting a valid text editor."""

OBJECT_PREVIEW_STRING_LENGTH = 30


def get_configuration() -> dict[str, str]:
    """Load the configuration file if available, with defaults for any
    keys not found. The config file should be "curifactory_config.json"
    in the project root.

    The defaults are:

    .. code-block:: json

        {
            "experiments_module_name": "experiments",
            "params_module_name": "params",
            "manager_cache_path": "data/",
            "cache_path": "data/cache",
            "runs_path": "data/runs",
            "logs_path": "logs/",
            "notebooks_path": "notebooks/",
            "reports_path": "reports/",
            "report_css_path": "reports/style.css",
        }

    Returns:
        the dictionary of configuration keys/values.
    """

    config_defaults = {
        "experiments_module_name": "experiments",
        "params_module_name": "params",
        "manager_cache_path": "data/",
        "cache_path": "data/cache",
        "runs_path": "data/runs",
        "logs_path": "logs/",
        "notebooks_path": "notebooks/",
        "reports_path": "reports/",
        "report_css_path": "reports/style.css",
    }

    # try to find configuration file in this dir or parent dirs (
    search_depth = 3
    prefix = ""
    while not os.path.exists(f"{prefix}{CONFIGURATION_FILE}") and search_depth > 0:
        prefix += "../"
        search_depth -= 1

    # get config file
    if os.path.exists(f"{prefix}{CONFIGURATION_FILE}"):
        with open(f"{prefix}{CONFIGURATION_FILE}") as infile:
            config = json.load(infile)

        # in case of any values that don't exist in explicit config
        for key in config_defaults:
            if key not in config:
                config[key] = config_defaults[key]

        # update paths if in subdir
        # NOTE: this doesn't update module names, because that requires actual code changes. Ideally,
        # this is only intended for things like notebooks to be able to use regular paths for live
        # curifactory usage, rather than actual experiment runs. For that it's on the user to correctly
        # ensure their current directory is set.
        for key in config:
            if key.endswith("_path"):
                config[key] = f"{prefix}{config[key]}"
    else:
        # defaults
        config = config_defaults

    return config


def get_editor() -> str:
    """Returns a text editor to use for richer text entry, such as in providing experiment notes.

    e.g. ``"vim"``
    """

    def first_available_editor():
        for editor_name in EDITORS:
            if shutil.which(editor_name) is not None:
                return editor_name
        return None

    editor = None
    if "EDITOR" in os.environ:
        editor = os.getenv("EDITOR")
    elif first_available_editor() is not None:
        editor = first_available_editor()
    elif os.name == "nt":
        editor = "notepad"

    if editor is None:
        raise RuntimeError("No valid text editor could be found.")
    return editor


def human_readable_mem_usage(byte_count: int) -> str:
    """Takes the given byte count and returns a nicely formatted string that includes the suffix (K/M/GB).

    Args:
        byte_count (int): The number of bytes to convert into KB/MB/GB.
    """

    negative = False
    if byte_count < 0:
        negative = True
        byte_count *= -1

    suffix = "B"
    if byte_count > 10**9:
        suffix = "GB"
        byte_count /= 10**9
    elif byte_count > 10**6:
        suffix = "MB"
        byte_count /= 10**6
    elif byte_count > 10**3:
        suffix = "KB"
        byte_count /= 10**3

    if negative:
        return f"-{byte_count:.2f}{suffix}"
    return f"{byte_count:.2f}{suffix}"


def human_readable_time(seconds: float) -> str:
    """Takes the given time in seconds and returns a nicely formatted string that includes the suffix.

    Args:
        seconds (float): The time in seconds to convert.
    """

    converted = seconds
    suffix = "s"

    # .1 = 100ms
    # .1ms = 100us = .0001
    # .1us = 100ns = .0000001

    if seconds > 60 * 60:
        suffix = "h"
        converted /= 60 * 60
    elif seconds > 60:
        suffix = "m"
        converted /= 60
    elif seconds < 0.0000001:
        suffix = "ns"
        converted *= 10**9
    elif seconds < 0.0001:
        suffix = "us"
        converted *= 10**6
    elif seconds < 0.1:
        suffix = "ms"
        converted *= 10**3

    return f"{converted:.2f}{suffix}"


def get_command_output(cmd: list[str], silent: bool = False) -> str:
    """Runs the command passed and returns the full string output of the command
    (minus the final newline).

    Args:
        cmd: Either a string command or array of strings, as one would pass to
            ``subprocess.run()``
    """
    try:
        cmd_return = subprocess.run(cmd, stdout=subprocess.PIPE)
    except:  # noqa: E722 -- TODO: we should actually handle this for the below note
        # NOTE - seems like we get filenotfound exceptions if it's an invalid command?
        # (e.g. on windows calling git when git isn't found from the env?)
        if not silent:
            logging.warning("Unable to run command '%s'" % cmd)
        return ""
    if cmd_return.returncode == 0:
        output = cmd_return.stdout.decode("utf-8")
        return output[:-2]  # TODO: (3/29/2023) is this definitely right?
        # seems like it would prob be different on windows vs linux
    return ""


def run_command(cmd: list[str]):
    """Prints output from running a command as it occurs.

    Args:
        cmd: Either a string command or array of strings, as one would pass to
            ``subprocess.run()``
    """
    logging.debug("Running command '%s'" % str(cmd))

    print(*cmd)

    with subprocess.Popen(cmd, stdout=subprocess.PIPE, bufsize=1, text=True) as p:
        for line in p.stdout:
            print(line, end="")  # process line here


def get_current_commit() -> str:
    """Returns printed output from running ``git rev-parse HEAD`` command."""
    return get_command_output(["git", "rev-parse", "HEAD"])


def check_git_dirty_workingdir() -> str:
    """Checks if git working directory is dirty or not. This is used to indicate
    potential reproducibility problems in the report and console output."""
    if get_command_output(["git", "diff", "--stat"]) != "":
        return True
    return False


def get_pip_freeze() -> str:
    """Returns printed output from running ``pip freeze`` command."""
    return get_command_output(["pip", "freeze"])


def get_conda_env() -> str:
    """Returns printed output from running ``conda env export --from-history`` command."""
    output = ""
    execs = ["conda", "mamba", "micromamba"]
    while output == "" and len(execs) > 0:
        cmd_word = execs[0]
        output = get_command_output([cmd_word, "env", "export", "--from-history"], True)
        execs.remove(cmd_word)
    if output == "":
        logging.warning("Unable to run conda or similar command")
        return output
    else:
        # fix random garbage at end of output (?!)
        output = output[:-7]
        return output


def get_conda_list() -> str:
    """Returns printed output from running ``conda env export --from-history`` command."""
    return get_command_output(["conda", "list"])


def get_os() -> str:
    """Get the current OS name and version."""
    return str(platform.platform())


def get_py_opening_comment(lines: list[str]) -> str:
    """Parse the passed lines of a python script file for a top of file docstring.
    This is used to get parameter and experiment file descriptions, so be sure to
    always comment your code!"""
    content = ""

    def check_for_end(line):
        if '"""' in line or "'''" in line:
            return True
        return False

    def endline_index(line):
        if '"""' in line:
            return line.index('"""')
        else:
            return line.index("'''")

    if len(lines[0]) > 3 and (lines[0].startswith('"""') or lines[0].startswith("'''")):
        if len(lines[0]) > 3 and not check_for_end(lines[0][3:]):
            content += lines[0][3:]

            for line in lines[1:]:
                if not check_for_end(line):
                    content += " " + line
                else:
                    content += " " + line[: endline_index(line)]
                    break
        elif len(lines[0]) > 3 and check_for_end(lines[0][3:]):
            content += lines[0][3 : endline_index(lines[0][3:]) + 3]

    # fix line breaks
    content = content.replace("\n", " ")
    while "  " in content:
        content = content.replace("  ", " ")

    return content


def set_logging_prefix(prefix: str):
    """Set the prefix content of the logger, which is incorporated in the log formatter. Currently
    this is just used to include pid number when running in parallel."""
    # https://stackoverflow.com/questions/17558552/how-do-i-add-custom-field-to-python-log-format-string
    old_factory = logging.getLogRecordFactory()

    if isinstance(old_factory, PrefixedLogFactory):
        old_factory.prefix = prefix
    else:
        logging.setLogRecordFactory(PrefixedLogFactory(old_factory, prefix))


class PrefixedLogFactory:
    """Note that we have to use this to prevent weird recursion issues.

    My understanding is that since logging is in the global context, after many many tests,
    the old_factory keeps getting set to the previous new_factory, and you end up with a massive
    function chain. Using this class approach above, we can check if we've already set the
    factory to an instance of this class, and just update the prefix on it.

    https://stackoverflow.com/questions/59585861/using-logrecordfactory-in-python-to-add-custom-fields-for-logging
    """

    def __init__(self, original_factory, prefix):
        self.original_factory = original_factory
        self.prefix = prefix

    def __call__(self, *args, **kwargs):
        record = self.original_factory(*args, **kwargs)
        record.prefix = self.prefix
        return record


def preview_object(object: any) -> str:
    """Get a small string representation of an object that fits nicely
    in a single line and includes shape information if relevant."""
    preview = f"({type(object).__name__}) {str(object)[:OBJECT_PREVIEW_STRING_LENGTH]}"
    if len(str(object)) > OBJECT_PREVIEW_STRING_LENGTH:
        preview += "..."
    if hasattr(object, "shape"):
        shape = None
        if callable(getattr(object, "shape", None)):
            shape = object.shape()
        else:
            shape = object.shape
        preview += f" shape: {shape}"
    elif hasattr(object, "__len__"):
        preview += f" len: {len(object)}"
    return preview


def init_logging(
    log_path=None,
    level=logging.INFO,
    log_errors=False,
    include_process=False,
    no_color=False,
    quiet=False,
    plain=False,
):
    """Sets up logging configuration, including the associated file output.

    Args:
        log_path (str): Folder to store output logs in. If ``None``, only log
            to console.
        level: The logging level to output.
        log_errors (bool): Whether to include error messages in the log output or not.
        include_process (bool): Whether to include the PID prefix value in the logger.
            This is mostly only used for when the ``--parallel`` flag is used, to
            help track which log message is from which process.
        no_color (bool): Suppress colors in console output.
        quiet (bool): Suppress all console log output.
        plain (bool): Output plain text log rather than rich output.
    """
    if include_process:
        # NOTE: taking out %(filename)s because it takes up space and makes the beginning of the log entries "jagged"
        plain_log_formatter = logging.Formatter(
            "%(asctime)s [%(levelname)s] {PID:%(process)s} - %(prefix)s%(message)s"
        )
        rich_log_formatter = logging.Formatter(
            "{PID:%(process)s} - %(prefix)s%(message)s"
        )
    else:
        plain_log_formatter = logging.Formatter(
            "%(asctime)s [%(levelname)s] - %(prefix)s%(message)s"
        )
        rich_log_formatter = logging.Formatter("%(prefix)s%(message)s")
    root_logger = logging.getLogger()
    root_logger.setLevel(level)
    root_logger.propagate = False
    root_logger.handlers = []

    if plain:
        # 4 characters so that it lines up all nice
        logging.addLevelName(logging.DEBUG, "DBUG")

    set_logging_prefix("")

    if log_path is not None:
        file_handler = logging.FileHandler(log_path)
        file_handler.setFormatter(plain_log_formatter)
        root_logger.addHandler(file_handler)

    if plain and not quiet:
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setFormatter(plain_log_formatter)
        root_logger.addHandler(console_handler)

    if not plain:
        if no_color:
            reconfigure(no_color=True)
        if not quiet:
            # console = Console(no_color=no_color)
            console_handler = RichHandler(
                console=get_console(),
                show_time=True,
                show_level=True,
                show_path=True,
                rich_tracebacks=True,
                tracebacks_show_locals=True,
                log_time_format="%X",
                keywords=["-----", "(aggregate)"],
            )
            console_handler.setFormatter(rich_log_formatter)
            root_logger.addHandler(console_handler)

    # https://stackoverflow.com/questions/27538879/how-to-disable-loggers-from-other-modules
    # disable all loggers except ours (TODO: may want to config this at some point)
    for name, logger in logging.root.manager.loggerDict.items():
        logger.disabled = True

    # sys.stdout = StreamToLogger(logging.INFO)
    if log_errors:
        sys.stderr = StreamToLogger(logging.ERROR)

    if quiet:
        root_logger.setLevel(logging.ERROR)


# https://stackoverflow.com/questions/19425736/how-to-redirect-stdout-and-stderr-to-logger-in-python
class StreamToLogger:
    def __init__(self, level):
        # self.logger = logger
        self.level = level

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            logging.log(self.level, line.rstrip())


def _warn_deprecation(msg):
    warnings.simplefilter("always", DeprecationWarning)  # turn off filter
    warnings.warn(msg, DeprecationWarning, stacklevel=3)
    warnings.simplefilter("default", DeprecationWarning)  # reset filter


class _DeprecationHelper:
    """Use this to warn about deprecated/changed class names.

    NOTE: this doesn't work for dataclasses, just make the old dataclass inherit from the new dataclass and throw the warning in __post_init__ and __init_subclass__

    See https://stackoverflow.com/questions/9008444/how-to-warn-about-class-name-deprecation
    """

    def __init__(self, original_name, new_target, extra_msg=""):
        self.original_name = original_name
        self.new_target = new_target
        self.extra_msg = extra_msg

    def _warn(self):
        _warn_deprecation(
            f"{self.original_name} has been deprecated. Please use {self.new_target.__class__} instead. {self.extra_msg}"
        )

    def __call__(self, *args, **kwargs):
        self._warn()
        return self.new_target(*args, **kwargs)

    def __getattr__(self, attr):
        self._warn()
        return getattr(self.new_target, attr)



"""This is the 'meta' curifactory CLI, which is a runnable to help set up a
project structure.

This file contains a ``__name__ == "__main__"`` and can be run directly.
"""

import argparse
import importlib.resources
import json
import os
import shutil

import curifactory as cf
from curifactory import utils


def initialize_project():  # noqa: C901 yeaaaah break up into sub functions
    print("Initializing curifactory project...")
    print(
        "Enter paths for configuration file, default values shown in '[]' (leave entry blank to use default)"
    )
    if os.path.exists(utils.CONFIGURATION_FILE):
        print(
            "NOTE: existing curifactory configuration found, default values are now currently existing values."
        )
    config = utils.get_configuration()

    def query_value_and_update(config, query, key):
        value = input(f"{query} ['{config[key]}']: ")
        if value != "":
            config[key] = value

    query_value_and_update(config, "Experiments module name", "experiments_module_name")
    query_value_and_update(config, "Parameters module name", "params_module_name")
    query_value_and_update(config, "Data path", "manager_cache_path")
    config["cache_path"] = os.path.join(config["manager_cache_path"], "cache/")
    config["runs_path"] = os.path.join(config["manager_cache_path"], "runs/")
    query_value_and_update(config, "Logs path", "logs_path")
    query_value_and_update(config, "Notebooks path", "notebooks_path")
    query_value_and_update(config, "Reports path", "reports_path")
    config["report_css_path"] = os.path.join(config["reports_path"], "style.css")

    # write out config
    print("Writing config file...")
    with open("curifactory_config.json", "w") as outfile:
        json.dump(config, outfile, indent=4)

    # create directory structure
    print("Setting up folders...")
    os.makedirs(config["experiments_module_name"], exist_ok=True)
    os.makedirs(config["params_module_name"], exist_ok=True)
    os.makedirs(config["manager_cache_path"], exist_ok=True)
    os.makedirs(config["cache_path"], exist_ok=True)
    os.makedirs(config["runs_path"], exist_ok=True)
    os.makedirs(config["logs_path"], exist_ok=True)
    os.makedirs(config["notebooks_path"], exist_ok=True)
    os.makedirs(os.path.join(config["notebooks_path"], "experiments"), exist_ok=True)
    os.makedirs(config["reports_path"], exist_ok=True)

    # copy in style sheet for reports
    # style_path = pkg_resources.resource_filename("curifactory", "data/style.css")
    with importlib.resources.as_file(
        importlib.resources.files("curifactory") / "data/style.css"
    ) as style_path:
        shutil.copyfile(style_path, config["report_css_path"])

    # copy in debug.py for easier IDE debugging entrypoint
    with importlib.resources.as_file(
        importlib.resources.files("curifactory") / "data/debug.py"
    ) as debug_file_path:
        shutil.copyfile(debug_file_path, "debug.py")

    # handle docker folder and dockerfile
    valid_docker_choice = False
    while not valid_docker_choice:
        docker_yn = input("Include docker folder and default dockerfile? [Y/n] ")
        if docker_yn == "" or docker_yn.lower() == "y":
            valid_docker_choice = True
            print("Setting up docker folder...")
            os.makedirs("docker", exist_ok=True)

            # read in the dockerfile contents and edit appropriately
            with importlib.resources.as_file(
                importlib.resources.files("curifactory") / "data/dockerfile"
            ) as dockerfile_path:
                with open(dockerfile_path) as infile:
                    contents = infile.read()
                    contents.replace("{{CF_VERSION}}", cf.__version__)
                with open("docker/dockerfile", "w") as outfile:
                    outfile.write(contents)

            with importlib.resources.as_file(
                importlib.resources.files("curifactory") / "data/startup.sh"
            ) as dockerfile_start_path:
                shutil.copyfile(dockerfile_start_path, "docker/startup.sh")

            with importlib.resources.as_file(
                importlib.resources.files("curifactory") / "data/.dockerignore"
            ) as dockerfile_ignore_path:
                shutil.copyfile(dockerfile_ignore_path, "docker/.dockerignore")
        elif docker_yn.lower() == "n":
            valid_docker_choice = True
        if not valid_docker_choice:
            print("Invalid entry, please enter 'y' or 'n', or leave blank for default.")

    # if this isn't being run in a git repository, ask the user if they want to git init
    if not os.path.exists(".git"):
        print(
            "No .git folder found. Curifactoy expects to run from within a git repository."
        )
        valid_gitinit_choice = False
        while not valid_gitinit_choice:
            gitinit_yn = input("Run `git init`? [y/N]")
            if gitinit_yn.lower() == "y":
                valid_gitinit_choice = True
                utils.run_command(["git", "init"])
            elif gitinit_yn == "" or gitinit_yn.lower() == "n":
                valid_gitinit_choice = True
            if not valid_gitinit_choice:
                print(
                    "Invalid entry, please enter 'y' or 'n', or leave blank for default."
                )

    # handle gitignore
    valid_gitfile_choice = False
    while not valid_gitfile_choice:
        gitfile_yn = input("Append curifactory paths to .gitignore? [Y/n] ")
        if gitfile_yn == "" or gitfile_yn.lower() == "y":
            valid_gitfile_choice = True

            newline_needed = False
            if os.path.exists(".gitignore"):
                with open(".gitignore") as infile:
                    lines = infile.readlines()
                    if len(lines) > 0 and lines[-1] != "\n":
                        newline_needed = True

            print("Appending to .gitignore...")
            with open(".gitignore", "a") as outfile:
                if newline_needed:
                    outfile.write("\n")
                outfile.write(
                    "\n".join(
                        [
                            "# curifactory paths",
                            os.path.join(config["manager_cache_path"], "*"),
                            os.path.join(config["cache_path"], "*"),
                            os.path.join(config["runs_path"], "*"),
                            os.path.join(config["logs_path"], "*"),
                            os.path.join(config["reports_path"], "*"),
                            "!" + config["report_css_path"],
                        ]
                    )
                )
        elif gitfile_yn.lower() == "n":
            valid_gitfile_choice = True
        if not valid_gitfile_choice:
            print("Invalid entry, please enter 'y' or 'n', or leave blank for default.")

    print("Curifactory project initialization complete!")
    add_completion_to_rc(False, False)


def add_completion_to_rc(bash=False, zsh=False):
    print(
        "\nFor bash/zsh completion to work, the argcomplete package needs to be installed (outside a conda environment)"
    )
    print("You can use `pip install argcomplete`")

    print("\nEnabling completion can then be done in three different ways:")
    print(
        "* Globally for all argcomplete python packages: \n\tsudo activate-global-python-argcomplete"
    )
    print(
        '* Single shell instance: \n\teval "$(register-python-argcomplete experiment)"'
    )
    print(
        "* Permanent non-globally for curifactory (RECOMMENDED):\n\tcurifactory completion --bash --zsh\n\tOr simply add single shell instance manually to any sourced shell rc file."
    )

    shell_string = '# tab-completion hook for curifactory\neval "$(register-python-argcomplete experiment)"'
    print(f"\n{shell_string}\n")

    if bash:
        print("Writing tab-completion hook into '~/.bashrc'...")
        with open(os.path.expanduser("~/.bashrc"), "a") as bashrc:
            bashrc.write(f"\n{shell_string}\n")
    if zsh:
        print("Writing tab-completion hook into '~/.zshrc'...")
        with open(os.path.expanduser("~/.zshrc"), "a") as zshrc:
            zshrc.write(f"\n{shell_string}\n")


def main():
    # TODO: could add the reset functionality as requested in this runnable.

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--version",
        dest="version",
        action="store_true",
        help="Print the current version of the curifactory library.",
    )

    subparsers = parser.add_subparsers(dest="subparser_name")

    init_parser = subparsers.add_parser("init")  # noqa: F841 -- we might use eventually

    completion_parser = subparsers.add_parser("completion")
    completion_parser.add_argument("--bash", dest="bash", action="store_true")
    completion_parser.add_argument("--zsh", dest="zsh", action="store_true")

    args = parser.parse_args()

    if args.version:
        print(cf.__version__)
        quit()

    if args.subparser_name == "init":
        initialize_project()
    elif args.subparser_name == "completion":
        add_completion_to_rc(args.bash, args.zsh)


if __name__ == "__main__":
    main()



"""Ensure that experiment pre-execution maps experiments as expected."""

# tests spec below
"""Ensure that an aggregate stage map correctly has the record references that went into it."""

"""Ensure that cache detection works for regular stage maps."""



"""Classes for managing an experiment's map/DAG logic."""

# NOTE: the "tree" concept is a little confusing linguistically in this file
# because the graph is being looked at both forwards and backwards. A "leaf
# stage" is a stage at the _end_ of the experiment graph, where its outputs are
# not needed as inputs in any other stages. However, for the DAG and treating
# the stage as a node, the leaf stages are actually "root nodes" for the
# execution trees, because we model all of a node's/stage's dependencies as its
# children/its subtree.

# It's also important to note that obviously trees are subsets of possible
# graphs, the DAG is not literally represented as the full actual graph, but a
# collection of the aforementioned execution trees. This is because for the
# purposes of the DAG we only care about this recursive backwards dependency
# pass. This means there's very possibly duplication across trees, in the case
# of nodes with outputs that are used in multiple other stages, but size isn't
# really going to be an issue, and (I think) it's much simpler to
# programatically construct and analyze it this way.


from curifactory.record import MapArtifactRepresentation, Record


class ExecutionNode:
    """Represents a particular stage for a particular record - a node in the
    overall experiment graph.

    Args:
        record (Record): The record in which this stage would execute.
        stage_name (str): The name of the stage that would execute.
    """

    def __init__(self, record: Record, stage_name: str):
        self.record = record
        self.stage_name = stage_name
        self.parent: ExecutionNode = None
        """The parent is a node that depends on this one/uses its outputs.
        (This means that the same execution node might appear in multiple
        execution trees, if its output is used by more than one other stage)"""

        self.dependencies: list[ExecutionNode] = []
        """Dependencies are the 'subtree' - the other nodes/stages that create
        the outputs that match this node's inputs."""

    def chain_rep(self) -> tuple[int, str]:
        """Return this node represented as a tuple of the record index and stage name."""
        return (self.record.get_record_index(True), self.stage_name)

    def __str__(self):
        return self.string_rep()

    def string_rep(self, level=0) -> str:
        """Recursively collect and return this node's index and name and that of
        its subtrees."""
        string = "\n"
        string += "  " * level
        string += f"({self.record.get_record_index(True)}, {self.stage_name})"
        for child in self.dependencies:
            string += child.string_rep(level + 1)
        return string


class DAG:
    """A DAG represents an entire mapped version of an experiment as a graph, where
    the nodes are stages and the connections are the outputs of one stage mapped to
    the associated inputs of another.

    The DAG is constructed as the first step of an experiment run (provided it isn't
    run with ``no_dag=True`` or ``--no-dag`` on the CLI), by setting the artifact
    manager to a special ``map_mode``. The experiment code is all run but every stage
    short circuits before execution and after collecting information about it
    (the record it's part of, what outputs are cached, etc.) This information is then
    used to determine which stages actually need to execute, working backwards from
    the leaf stages. This differs from curifactory's base ``no_dag`` mode because the
    need-to-execute for every stage is based primarily on whether any future stage
    actually requires this one's outputs and has a need-to-execute (resulting in a
    recursive check backwards from the leaf stages.)
    """

    def __init__(self):
        self.records: list[Record] = []
        self.artifacts: list[MapArtifactRepresentation] = []
        """This should essentially be an equivalent copy of ArtifactManager.artifacts.
        All of record's stage inputs and outputs should correctly index into this."""

        self.execution_list: list[tuple[int, str]] = []
        """This is the list of ExecutionNode individual (non-recursive) string
        representations: ``(RECORD_ID, STAGE_NAME)``"""

        self.execution_trees: list[ExecutionNode] = []
        """The set of node execution trees - each node here is a "leaf stage", or
        stage with no outputs that other stages depend on. This is essentially the
        inverted tree, because stage "leafs" will each be a root of an execution
        tree, where the sub-trees are all the dependencies required for it to run."""

    def analyze(self):
        """Construct execution trees and execution list."""
        self.build_execution_trees()
        self.determine_execution_list()

    def get_record_string(self, record_index: int) -> str:
        """Get a string representation for the given record. This collects
        all of the associated stages, inputs and outputs for each, and cache
        status for each artifact."""
        record = self.records[record_index]
        output = (
            f"==== {record.get_reference_name(True)} hash: {record.get_hash()} ===="
        )
        for index, stage in enumerate(record.stages):
            output += "\nStage: " + stage
            if self.is_leaf(record, stage):
                output += " (leaf)"
            if len(record.input_records) > 0:
                output += "\n\tInput records:"
                for input_record in record.input_records:
                    output += f"\n\t\t{input_record.get_reference_name(True)}"
            if len(record.stage_inputs[index]) > 0:
                output += "\n\tInputs:"
                for stage_input_index in record.stage_inputs[index]:
                    if stage_input_index != -1:
                        stage_input = self.artifacts[stage_input_index]
                        output += f"\n\t\t{stage_input.name}"
                        if stage_input.cached:
                            output += f" (cached) [{stage_input.metadata['manager_run_info']['reference']}]"
            if len(record.stage_outputs[index]) > 0:
                output += "\n\tOutputs:"
                for stage_output_index in record.stage_outputs[index]:
                    stage_output = self.artifacts[stage_output_index]
                    output += f"\n\t\t{stage_output.name}"
                    if stage_output.cached:
                        output += f" (cached) [{stage_output.metadata['manager_run_info']['reference']}]"
        return output

    def print_experiment_map(self):
        """Print the representations for each record."""
        string = ""
        for index, record in enumerate(self.records):
            string += self.get_record_string(index) + "\n"
        print(string)

    def is_leaf(self, record: Record, stage_name: str) -> bool:
        """Check if the given stage is a leaf, based on two conditions:

        1. Stage is a leaf if it has no output artifacts.
        2. Stage is a leaf if it has outputs but they aren't used as inputs in
            any other stages.
        """
        stage_index = record.stages.index(stage_name)
        outputs = record.stage_outputs[stage_index]

        # it's a leaf if no outputs
        if len(outputs) == 0:
            return True

        # it's a leaf if outputs are not used anywhere
        for output_index in outputs:
            output = self.artifacts[output_index]
            if self.is_output_used_anywhere(record, stage_index + 1, output.name):
                return False
        return True

    def is_output_used_anywhere(
        self, record: Record, stage_search_start_index: int, output: str
    ) -> bool:
        """Check if the specified output is used as input in any stage."""
        # Iterate each following stage in that record and see if the requested output
        # is in any of the inputs
        # TODO: instead of checking by name, we should be checking by the artifact id
        for i in range(stage_search_start_index, len(record.stages)):
            # NOTE: this works for both stages _and_ aggregates because expected_state
            # reps for aggregates get added to record.stage_inputs in the aggregate decorator
            for stage_input_index in record.stage_inputs[i]:
                stage_input = self.artifacts[stage_input_index]
                if stage_input.name == output:
                    return True

        # check if any following records directly use the output in a stage or aggregate
        children = self.child_records(record)
        for child in children:
            # check if we're directly using it in a normal stage
            explicitly_used = self.is_output_used_anywhere(child, 0, output)
            if explicitly_used:
                return True

        return False

    def child_records(self, record: Record) -> list[Record]:
        """Return a list of all records for which the provided record is an input record.
        (This occurs when calling ``record.make_copy()`` and for aggregates.)"""
        children = [
            other_record
            for other_record in self.records
            if record in other_record.input_records
        ]
        return children

    def find_leaves(self) -> list[tuple[int, str]]:
        """Get all of the nodes who have no outputs depended on by any others, these
        should be all of the "last" stages in the experiment and/or utility stages
        (e.g. a stage that just handles reporting or something like that/doesn't really
        output any artifacts.)

        Returns:
            a list of tuples where the first element is the record index and the
            second is the name of the stage.
        """
        leaves = []
        for index, record in enumerate(self.records):
            for stage in record.stages:
                if self.is_leaf(record, stage):
                    leaves.append((index, stage))

        return leaves

    def build_execution_tree_recursive(
        self, record: Record, stage: str
    ) -> ExecutionNode:
        """Recursively builds the stage dependency tree based on inputs/outputs. This does
        not condition anything based on cache or overwrite status, this is exclusively the
        "inverted" stage path (provided stage is the root)."""
        this_node = ExecutionNode(record, stage)

        # go through each input and get the record and stage that provides it as an output.
        # These are the dependencies, so recursively create nodes for them
        stage_index = record.stages.index(stage)
        for literal_input_index, stage_input_index in enumerate(
            record.stage_inputs[stage_index]
        ):
            # NOTE: literal_input_index is just the for loop index so we can index
            # into the associated string input names list on the record

            # Did we not find the input?
            if stage_input_index == -1:
                # if the stage's suppress missing inputs is set, just continue
                if record.stage_suppress_missing[stage_index]:
                    continue

                # if kwargs passed this input name, just continue
                if (
                    record.stage_inputs_names[stage_index][literal_input_index]
                    in record.stage_kwargs_keys[stage_index]
                ):
                    continue

                # otherwise throw a KeyError (see 260 staging.py)
                raise KeyError(
                    "Stage '%s' input '%s' not found in record state and not passed to function call. Set 'suppress_missing_inputs=True' on the stage and give a default value in the function signature if this should run anyway."
                    % (
                        stage,
                        record.stage_inputs_names[stage_index][literal_input_index],
                    )
                )

            stage_input: MapArtifactRepresentation = self.artifacts[stage_input_index]

            prereq_record = self.records[stage_input.record_index]
            prereq_stage = stage_input.stage_name

            sub_node = self.build_execution_tree_recursive(prereq_record, prereq_stage)
            sub_node.parent = this_node
            this_node.dependencies.append(sub_node)

        return this_node

    def build_execution_trees(self):
        """Build an execution tree (essentially the sub-DAG) for every leaf node found."""
        self.execution_trees = []
        leaves = self.find_leaves()
        for leaf in leaves:
            leaf_record = self.records[leaf[0]]
            leaf_stage = leaf[1]
            tree = self.build_execution_tree_recursive(leaf_record, leaf_stage)
            self.execution_trees.append(tree)

    def determine_execution_list(self):
        """I've got them on the list, they'll none of them be missed."""
        self.execution_list = []
        for node in self.execution_trees:
            self.determine_execution_list_recursive(node, False)

    def determine_execution_list_recursive(
        self, node: ExecutionNode, overwrite_check_only: False
    ) -> bool:
        """Determines if the requested stage will need to execute or not, and if so prepends
        itself and all prior stages needed to execute through recursive calls."""
        stage_index = node.record.stages.index(node.stage_name)

        # first check if all of the outputs for this stage are cached, if any one of them is not,
        # will need to add to execution chain
        cached = True
        for stage_output_index in node.record.stage_outputs[stage_index]:
            stage_output: MapArtifactRepresentation = self.artifacts[stage_output_index]
            if not stage_output.cached:
                cached = False
                break

        # there is a special case here where a stage has _no_ outputs. Since the
        # above loop is assuming cached by default, and we have no non-cached things
        # to prove otherwise, we will hit this point in the code with cached=True
        # we want stages with no outputs to always execute, so we check for that and
        # force cached False
        if len(node.record.stage_outputs[stage_index]) == 0:
            cached = False

        # -- overwrite seek mode --
        # (in this mode we add the node if and only if there's a sub node that's being overwritten.)

        overwrite_stage_found = False
        """set to true if either this stage is set to overwrite, or if something in the subtree is
        going to be overwritten"""

        if node.stage_name in node.record.manager.overwrite_stages:
            overwrite_stage_found = True

        # we don't bother stepping into this part if _this_ stage is specified to overwrite,
        # because in the dependencies required mode below we'll be going on to check dependency
        # stages normally anyway.
        if (
            cached and not node.record.manager.overwrite and not overwrite_stage_found
        ) or overwrite_check_only:
            # recursively go through previous stages whose output is needed for this one and
            # look for a stage on the manager's overwrite stages list.
            for sub_node in node.dependencies:
                overwrite_found = self.determine_execution_list_recursive(
                    sub_node, True
                )
                if overwrite_found:
                    overwrite_stage_found = True
                    break

            # if no dependencies need to be overwritten, we're good to stop going down.
            if not overwrite_stage_found:
                return False
            # otherwise continue into dependencies required mode (because we need to execute
            # this stage)

        # -- dependencies required mode --
        # (we hit this mode because we know this stage needs to execute, so now recursively go
        # through its dependencies to see if _those_ need to execute as well.)

        # need to execute, off with its head!
        if node.chain_rep() not in self.execution_list:
            self.execution_list.insert(0, node.chain_rep())
            # NOTE: for some weird reason in how I have this structured, if an overwrite stage
            # is found, the ordering of the execution list is in reverse of what it's supposed to be.
            # For right now that doesn't really matter, but if we ever get fancy with using the DAG
            # to directly run functions, it will be important to fix this.

        # now recursively go through previous stages whose output is needed for this one and
        # continue to build execution list.
        for sub_node in node.dependencies:
            self.determine_execution_list_recursive(sub_node, False)

        # the output from this is technically only checked for inside the overwrite seek mode
        # I think
        return overwrite_stage_found



"""Testing the decorators to help orchestrate caching and
input/output passing through record state between stages."""

import copy
import logging
import os
import pickle
import time
from functools import wraps
from typing import Any, Union

import psutil

from curifactory import utils
from curifactory.caching import Cacheable, FileReferenceCacher, Lazy, PickleCacher
from curifactory.record import ArtifactRepresentation, MapArtifactRepresentation, Record

# NOTE: resource only exists on unix systems
if os.name != "nt":
    import resource


class InputSignatureError(Exception):
    pass


class OutputSignatureError(Exception):
    pass


class EmptyCachersError(Exception):
    pass


class ExecutingWithSkippedInputError(Exception):
    pass


class CachersMismatchError(Exception):
    pass


# TODO: this doesn't belong here but not sure where to put it
class SkippedOutput:
    """An object placed into record state when a stage's execution is skipped based
    on the DAG. This is so that the inputs check at the beginning of the stage (prior
    to DAG execution check) does not fail."""

    def __init__(
        self, record: Record, stage_name: str, artifact_rep: ArtifactRepresentation
    ):
        self.artifact = artifact_rep
        self.dag_rep = (record.get_record_index(), stage_name)


def _log_stats(
    record,
    pre_cache_time_start,
    pre_cache_time_end,
    pre_mem_usage,
    post_mem_usage,
    exec_time_start=0,
    exec_time_end=0,
    post_cache_time_start=0,
    post_cache_time_end=0,
    pre_max_footprint=0,
    post_max_footprint=0,
):
    pre_cache_time = pre_cache_time_end - pre_cache_time_start
    exec_time = exec_time_end - exec_time_start
    post_cache_time = post_cache_time_end - post_cache_time_start
    cache_time = pre_cache_time + post_cache_time

    mem_change = post_mem_usage - pre_mem_usage

    footprint_change = post_max_footprint - pre_max_footprint

    logging.debug(
        "Memory (current usage/max allocated) - %s / %s"
        % (
            utils.human_readable_mem_usage(post_mem_usage),
            utils.human_readable_mem_usage(post_max_footprint),
        )
    )
    logging.debug(
        "Stage memory impact (current/max) - %s / %s"
        % (
            utils.human_readable_mem_usage(mem_change),
            utils.human_readable_mem_usage(footprint_change),
        )
    )
    logging.debug(
        "Timing - execution: %s  caching: %s"
        % (utils.human_readable_time(exec_time), utils.human_readable_time(cache_time))
    )


def stage(  # noqa: C901 -- TODO: will be difficult to simplify...
    inputs: list[str] = None,
    outputs: list[Union[str, Lazy]] = None,
    cachers: list = None,
    suppress_missing_inputs: bool = False,
):
    """Decorator to wrap around a function that represents a single step in an experiment,
    a block with inputs and outputs pertaining to the remainder of that experiment.

    Important:
        Any function wrapped with the stage decorator must take a Record instance as the first
        parameter, followed by the input parameters corresponding to the :code:`inputs` list.

    Args:
        inputs (list[str]): A list of variable names that this stage will need from the
            record state. **Note that all inputs listed here must have a corresponding
            input argument in the function definition line, each with the exact same name
            as in this list.**
        outputs (list[Union[str, Lazy]]): A list of variable names that this stage will return and store
            in the record state. These represent, in order, the tuple of returned values from
            the function being wrapped.
        cachers (list[Cacheable]): An optional list of ``Cacheable`` instances ("strategies") to
            apply to each of the return outputs. If specified, for each output, an instance
            of the corresponding cacher is initialized, and the ``save()`` function is called.
            Before the wrapped function is called, the output path is first checked, and if it
            exists and the current record parameter set is not configured to overwrite, the ``load()`` function
            is called and the wrapped function **does not execute.** Note that caching is all
            or nothing for a single function, you cannot cache only one returned value out of
            several.
        suppress_missing_inputs (bool): If true, any stage inputs that are not found in the record's
            state will be passed in as ``None`` rather than raising an exception. This can
            be used to make all inputs optional, such as if a stage will be used after different
            sets of previous stages and not all values are necessarily required.

    Example:
        .. code-block:: python

            @stage(inputs=["data", "model"], outputs=["results"], cachers=[JsonCacher])
            def test_model(record: Record, data: pd.DataFrame, model):
                # ...
                return results_dictionary

        Note that from this example, this stage assumes some other stages have output
        ``"data"`` and ``"model"`` at some point.
    """

    def decorator(function):
        @wraps(function)
        def wrapper(record: Record, *args, **kwargs):
            # set the logging prefix to the parameter set name
            if record.params is not None:
                utils.set_logging_prefix(f"[{record.params.name}] ")
            else:
                utils.set_logging_prefix("")

            name = function.__name__
            if record.manager.map_mode:
                logging.debug("Mapping stage %s", name)
            else:
                logging.info("-----")
                logging.info("Stage %s", name)
            pre_footprint = 0
            if os.name != "nt":
                pre_footprint = (
                    resource.getrusage(resource.RUSAGE_SELF).ru_maxrss * 1024
                )
            pre_mem_usage = psutil.Process().memory_info().rss
            if record.manager.stage_active:
                logging.warn(
                    "Stage '%s' executed while another stage ('%s') was already running. Directly executing a stage from another stage is not advised."
                    % (name, record.manager.current_stage_name)
                )
            # apply consistent handling
            nonlocal inputs, outputs, cachers
            if inputs is None:
                inputs = []
            if outputs is None:
                outputs = []

            record.manager.current_stage_name = name
            record.manager.stage_active = True
            record.stages.append(name)
            record.stage_outputs.append([])
            record.stage_inputs.append([])
            record.stage_suppress_missing.append(suppress_missing_inputs)
            record.stage_kwargs_keys.append(list(kwargs.keys()))
            record.stage_inputs_names.append(inputs)
            record.manager.update_map_progress(record, "start")

            # check for lazy object / cacher mismatch in outputs
            for output in outputs:
                if (
                    type(output) == Lazy
                    and cachers is None
                    and not record.manager.lazy
                    and not record.manager.ignore_lazy
                ):
                    raise OutputSignatureError(
                        "Stage outputs for '%s' contain Lazy objects but do not have cachers specified. Please provide cachers for this stage."
                        % name
                    )

            # replace any non-lazy outputs with lazy or vice-versa depending on manager flags.
            if record.manager.lazy:
                no_cachers = False
                for index, output in enumerate(outputs):
                    if type(output) != Lazy:
                        logging.debug("Forcing lazy cache for '%s'" % output)
                        outputs[index] = Lazy(output)
                        # NOTE: since Lazy caching doesn't work without a cacher, we need to ensure
                        # one if none exists. Pickle is pretty broad, but obviously there are some things
                        # that don't work, so we need to warn about this
                        if cachers is None:
                            no_cachers = True
                            logging.warning(
                                "Stage %s does not have cachers specified, a --lazy run will force caching by applying PickleCachers to anything with none specified, but this can potentially cause errors."
                                % name
                            )
                            cachers = []
                        if no_cachers:
                            cachers.append(PickleCacher)
            elif record.manager.ignore_lazy:
                for index, output in enumerate(outputs):
                    if type(output) == Lazy:
                        logging.debug("Disabling lazy cache for '%s'" % output)
                        outputs[index] = output.name

            # check for mismatched amounts of cachers
            if cachers is not None and len(cachers) != len(outputs):
                raise CachersMismatchError(
                    f"Stage '{name}' - the number of cachers does not match the number of outputs to cache."
                )

            # find any required inputs for the function in the record
            # NOTE: this is before checking cache so that we can keep a record of what inputs we're expecting
            # (for mapping purposes)
            # note that we turn off lazy resolution because we haven't yet
            # determined if the stage even needs to run yet or not. We handle
            # resolution manually farther down.
            record.state.resolve = False
            function_inputs = {}
            for function_input in inputs:
                # NOTE: this means the state was probably manually changed
                if function_input not in record.state_artifact_reps:
                    record.stage_inputs[-1].append(-1)
                else:
                    record.stage_inputs[-1].append(
                        record.state_artifact_reps[function_input]
                    )
                # TODO: add None if function_input is not in state. (There are a few cases
                # where this shouldn't be a problem, e.g. if everything is cached but a new
                # input was added - this shouldn't inherently cause a problem unless the
                # output would be wrong.
                if function_input not in record.state:
                    if suppress_missing_inputs and not record.manager.map_mode:
                        logging.warning(
                            "Suppressed missing inputs, will expect function signature default value for '%s' or a direct argument pass on the stage function call..."
                            % function_input
                        )
                        # function_inputs[function_input] = None
                        # NOTE: we don't actually need to pass in None, we can expect the user to implement whatever defaults they want for optional parameters. If they don't specify a default, this will fail as normal.
                    elif function_input not in kwargs and not record.manager.map_mode:
                        raise KeyError(
                            "Stage '%s' input '%s' not found in record state and not passed to function call. Set 'suppress_missing_inputs=True' on the stage and give a default value in the function signature if this should run anyway."
                            % (name, function_input)
                        )
                else:
                    function_inputs[function_input] = record.state[function_input]
            function_inputs.update(kwargs)
            record.state.resolve = True

            # note to future self and anyone else who's IDE says this is repeated code (with aggregate below)
            # no, you cannot abstract this into _check_cached_outputs - if you try to reassign to cachers from
            # another function, because of the deep voodoo black magic sorcery that is decorators with arguments,
            # it considers it different code.
            if cachers is not None:
                # instantiate cachers if not already
                for i in range(len(cachers)):
                    cacher = cachers[i]
                    if type(cacher) == type:
                        cachers[i] = cacher()
                    # set the active record on the cacher as well as provide a default name
                    # (the name of the output)
                    cachers[i].set_record(record)
                    cachers[
                        i
                    ].stage = name  # set current stage name, so get_path is correct in later stages (particularly for lazy)
                    if cachers[i].name is None and cachers[i].path_override is None:
                        if type(outputs[i]) == Lazy:
                            cachers[i].name = outputs[i].name
                        else:
                            cachers[i].name = outputs[i]
            record.stage_cachers = cachers

            # at this point we've grabbed all information we would need if we're
            # just mapping out the stages, so return at this point.
            # TODO: 3/8/2023 - not true, we're not getting outputs (7/6/2023 is this still an issue? seems fine?)
            if record.manager.map_mode:
                record.manager.stage_active = False
                # in order to get a more detailed map that accurately shows input/output names,
                # we need to create pseudo-state-artifact-representations for the outputs. Since
                # we obviously can't add the actual artifacts (there are none without running!),
                # we just add the string key and a name, so record __repr__ has something to use
                _get_output_representations_for_map(record, outputs, cachers, None)
                record.stage_cachers = None
                return record

            # determine if we need to execute this stage and handle any
            # cached values.
            execute_stage = False  # false by default so the non-dag condition below can distinguish whether
            # the dag condition actually set execute stage to true or if just default
            pre_cache_time_start = time.perf_counter()  # time to load from cache
            record.manager.lock()

            # if we have an execution list from our stage DAG, use that
            # to determine if this stage executes or not.
            if record.manager.map is not None:
                stage_rep = (record.get_record_index(), name)
                if stage_rep not in record.manager.map.execution_list:
                    logging.debug('DAG-indicated stage skip "%s".' % str(stage_rep))
                    _dag_skip_check_cached_outputs(name, record, outputs, cachers)

                    # grab any possible previous reportables so they still end up in report.
                    _check_cached_reportables(name, record)
                    record.store_tracked_paths()
                    execute_stage = False
                else:
                    # the representation is in the execution list, so execute!
                    execute_stage = True

            # otherwise, proceed normally with cache/load checks
            # NOTE: this is an explicit _separate_ check because if our DAG indicates that
            # this stage executes, we still want to actually double check the cache and determine
            # if we _still_ need to execute this stage. This is primarily for cases where a
            # stage happens to get run multiple times with different records that have the same args/
            # same outputs. The DAG doesn't dynamically update with cached values during the actual
            # experiment run, so it won't catch this case by itself.
            # So the flow is: if we have a DAG, and it says to execute this stage, or if we don't have
            # a DAG, check if we actually need to run this based on cached values.
            if record.manager.map is None or execute_stage:
                cache_valid = _check_cached_outputs(name, record, outputs, cachers)
                if cache_valid:
                    # get previous reportables if available
                    _check_cached_reportables(name, record)

                    # if we've hit this point, we will be returning early/not executing
                    # the stage because all outputs are found. The process of checking
                    # cached outputs should correctly add all the necessary tracked paths,
                    # so to transfer these paths into a store full run, we just need
                    # the below call
                    # NOTE: I _believe_ that we also get metadata from this because
                    # the load_metadata will have entered the metadata path already.
                    # this will need to be tested
                    record.store_tracked_paths()
                    execute_stage = False
                else:
                    # at least one output wasn't cached, so execute order 66!
                    execute_stage = True

            # check each input for Lazy objects and load them if we know we have to execute this stage
            if execute_stage:
                for function_input in function_inputs:
                    if (
                        type(function_inputs[function_input]) == Lazy
                        and function_inputs[function_input].resolve
                    ):
                        logging.debug(
                            "Resolving lazy load object '%s'" % function_input
                        )
                        function_inputs[function_input] = function_inputs[
                            function_input
                        ].load()
                    # make sure a skipped output hasn't made it through - this would indicate
                    # a mistake in the DAG or perhaps the user didn't correctly list all expected
                    # state inputs in an aggregate
                    if type(function_inputs[function_input]) == SkippedOutput:
                        raise ExecutingWithSkippedInputError(
                            "Input '%s' was never computed, indicating a DAG error. Try running with '--no-map'."
                            % function_input
                        )

            record.manager.unlock()
            pre_cache_time_end = time.perf_counter()
            if not execute_stage:
                post_mem_usage = psutil.Process().memory_info().rss
                post_footprint = 0
                if os.name != "nt":
                    post_footprint = (
                        resource.getrusage(resource.RUSAGE_SELF).ru_maxrss * 1024
                    )
                _log_stats(
                    record,
                    pre_cache_time_start,
                    pre_cache_time_end,
                    pre_mem_usage,
                    post_mem_usage,
                    0,
                    0,
                    0,
                    0,
                    pre_footprint,
                    post_footprint,
                )
                utils.set_logging_prefix("")
                record.manager.stage_active = False
                record.manager.update_map_progress(record, "continue")
                record.stage_cachers = None
                return record

            # run the function
            logging.info("Stage %s executing...", name)
            exec_time_start = time.perf_counter()
            try:
                function_outputs = function(record, *args, **function_inputs)
            except TypeError as e:
                raise InputSignatureError(
                    "Signature for '%s' does not match stage input list. Signature should include %s, or there may be missing default values for a stage called with suppress_missing_inputs. Sub error: %s"
                    % (name, str(inputs), str(e))
                )
            exec_time_end = time.perf_counter()

            # handle storing outputs in record
            post_cache_time_start = time.perf_counter()
            record.manager.lock()
            _store_outputs(name, record, outputs, cachers, function_outputs)
            _store_reportables(name, record)
            record.store_tracked_paths()
            record.manager.unlock()
            post_cache_time_end = time.perf_counter()

            # free up any memory from cached things
            cleaned_function_outputs = []
            lazy_found = False

            # iterate the outputs rather than function_outputs
            for index, output_name in enumerate(outputs):
                if type(output_name) == Lazy:
                    lazy_found = True
                    logging.debug("Lazy object '%s' will be cleaned." % output_name)
                    if index == 0 and len(outputs) == 1:
                        cleaned_function_outputs = outputs[
                            index
                        ]  # so that this doesn't syntactically read output_name, because it's Lazy not a str
                    else:
                        cleaned_function_outputs.append(outputs[index])
                else:
                    if index == 0 and len(outputs) == 1:
                        cleaned_function_outputs = function_outputs
                    else:
                        cleaned_function_outputs.append(function_outputs[index])
            if len(outputs) > 1:
                cleaned_function_outputs = tuple(cleaned_function_outputs)

            # free up any lazy cache objects
            if lazy_found:
                logging.debug("Freeing memory from lazy objects...")
                pre_del_mem_usage = psutil.Process().memory_info().rss
                del function_outputs
                post_del_mem_usage = psutil.Process().memory_info().rss
                del_mem_diff = pre_del_mem_usage - post_del_mem_usage
                logging.debug("Freed %s" % utils.human_readable_mem_usage(del_mem_diff))

            logging.info("Stage %s complete", name)

            # check memory usage
            post_footprint = 0
            if os.name != "nt":
                post_footprint = (
                    resource.getrusage(resource.RUSAGE_SELF).ru_maxrss * 1024
                )
            post_mem_usage = psutil.Process().memory_info().rss

            _log_stats(
                record,
                pre_cache_time_start,
                pre_cache_time_end,
                pre_mem_usage,
                post_mem_usage,
                exec_time_start,
                exec_time_end,
                post_cache_time_start,
                post_cache_time_end,
                pre_footprint,
                post_footprint,
            )

            record.output = cleaned_function_outputs
            utils.set_logging_prefix("")
            record.manager.stage_active = False
            record.manager.update_map_progress(record, "continue")
            record.stage_cachers = None
            return record

        return wrapper

    return decorator


def aggregate(  # noqa: C901 -- TODO: will be difficult to simplify...
    inputs: list[str] = None, outputs: list[str] = None, cachers: list = None
):
    """Decorator to wrap around a function that represents some step that must operate across
    multiple different parameter sets or "execution chains" within an experiment. This is normally
    used to run final analyses and comparisons of results across all passed parameter sets.

    Important:
        Any function wrapped with the aggregate decorator must take a Record instance as the first
        argument and a list of Record instances as the second. The former is the record that applies
        to this function, and the latter is the set of other records from elsewhere in the experiment
        that this function needs to aggregate across.

    Args:
        inputs (list[str]): A list of variable names this stage expects to find in the
            state of each record passed into it. A warning will be thrown on any records that
            do not have the requested variable in state. Variables listed here are used for the
            DAG/map calculation to determine which stages are actually required to run for this
            stage to have everything it needs. **Note that all inputs listed here must have a
            corresponding input parameter in the function definition line, each with the exact
            same name as in this list.** These arguments are each dictionaries of the requested
            state artifacts, keyed by the record they come from.
        outputs (list[str]): A list of variable names that this stage will return and store
            in the record state. These represent, in order, the tuple of returned values from
            the function being wrapped.
        cachers (list[Cacheable]): An optional list of ``Cacheable`` instances ("strategies") to
            apply to each of the return outputs. If specified, for each output, an instance
            of the corresponding cacher is initialized, and the ``save()`` function is called.
            Before the wrapped function is called, the output path is first checked, and if it
            exists and the current record parameters are not set to overwrite, the ``load()``
            function is called and the wrapped function **does not execute.** Note that caching
            is all or nothing for a single function, you cannot cache only one returned value
            out of several.

    Example:
        .. code-block:: python

            @aggregate(["results"], ["final_results"], [JsonCacher])
            def compile_results(record: Record, records: List[Record], results: dict[Record, float]):
                final_results = {}
                for in_record, result in results.items():
                    results[in_record.params.name] = result
                return final_results
    """

    def decorator(function):
        @wraps(function)
        def wrapper(record: Record, records: list[Record] = None, **kwargs):
            # set the logging prefix to the parameter set name
            if record.params is not None:
                utils.set_logging_prefix(f"[{record.params.name}] ")
            else:
                utils.set_logging_prefix("")

            if records is None:
                # if no explicit request for a specific set of records given,
                # use all previous records (except own record)
                records = [
                    manager_record
                    for manager_record in record.manager.records
                    if manager_record != record
                ]
            record.input_records = records

            name = function.__name__
            if record.manager.map_mode:
                logging.debug("Mapping aggregate stage %s", name)
            else:
                logging.info("-----")
                logging.info("Stage (aggregate) %s", name)
            pre_footprint = 0
            if os.name != "nt":
                pre_footprint = (
                    resource.getrusage(resource.RUSAGE_SELF).ru_maxrss * 1024
                )
            pre_mem_usage = psutil.Process().memory_info().rss
            if record.manager.stage_active:
                logging.warn(
                    "Stage '%s' executed while another stage ('%s') was already running. Directly executing a stage from another stage is not advised."
                    % (name, record.manager.current_stage_name)
                )

            # apply consistent handling
            nonlocal inputs, outputs, cachers
            if inputs is None:
                inputs = []
            if outputs is None:
                outputs = []

            record.manager.current_stage_name = name
            record.set_aggregate(records)
            record.stages.append(name)
            record.stage_outputs.append([])
            record.stage_inputs.append([])
            record.stage_suppress_missing.append(False)
            # TODO: (7/24/2023) wait is False above correct, this is largely irrelevant for aggregate
            # but unclear where exactly in dag mapping I'm going to be checking this.
            record.stage_kwargs_keys.append(list(kwargs.keys()))
            # TODO: (7/24/2023) if a user is passing kwargs to an aggregate, they'll need to make sure
            # they're actually passing a dictionary of records. Eventually we could prob just check that
            # here.
            record.stage_inputs_names.append(inputs)
            record.manager.update_map_progress(record, "start")

            # check for lazy object / cacher mismatch in outputs
            for output in outputs:
                if (
                    type(output) == Lazy
                    and cachers is None
                    and not record.manager.lazy
                    and not record.manager.ignore_lazy
                ):
                    raise OutputSignatureError(
                        "Stage outputs for '%s' contain Lazy objects but do not have cachers specified. Please provide cachers for this stage."
                        % name
                    )

            # replace any non-lazy outputs with lazy or vice-versa depending on manager flags.
            if record.manager.lazy:
                no_cachers = False
                for index, output in enumerate(outputs):
                    if type(output) != Lazy:
                        logging.debug("Forcing lazy cache for '%s'" % output)
                        outputs[index] = Lazy(output)
                        # NOTE: since Lazy caching doesn't work without a cacher, we need to ensure
                        # one if none exists. Pickle is pretty broad, but obviously there are some things
                        # that don't work, so we need to warn about this
                        if cachers is None:
                            no_cachers = True
                            logging.warning(
                                "Aggregate stage %s does not have cachers specified, a --lazy run will force caching by applying PickleCachers to anything with none specified, but this can potentially cause errors."
                                % name
                            )
                            cachers = []
                        if no_cachers:
                            cachers.append(PickleCacher)
            elif record.manager.ignore_lazy:
                for index, output in enumerate(outputs):
                    if type(output) == Lazy:
                        logging.debug("Disabling lazy cache for '%s'" % output)
                        outputs[index] = output.name

            # check for mismatched amounts of cachers
            if cachers is not None and len(cachers) != len(outputs):
                raise CachersMismatchError(
                    f"Stage '{name}' - the number of cachers does not match the number of outputs to cache"
                )

            # similar to how inputs are checked on the state for a regular stage, we use
            # `inputs` to (softly) check each input record for the requested state
            # variables. This is useful both from a documentation standpoint and also
            # is required for the DAG computation to work as expected.
            #
            # For each requested artifact in inputs, we construct a dictionary where the
            # values are those artifacts from the input records, and keyed by the associated
            # records themselves.
            function_inputs = {}
            for function_input in inputs:
                function_inputs[function_input] = {}
                for prev_record in records:
                    if function_input not in prev_record.state_artifact_reps:
                        # TODO: (7/18/2023) unclear if this warning is still necessary or
                        # not, possibly just make this a regular info level?
                        logging.warning(
                            "Artifact '%s' not found in %s"
                            % (function_input, prev_record.get_reference_name())
                        )
                    else:
                        record.stage_inputs[-1].append(
                            prev_record.state_artifact_reps[function_input]
                        )

                    # NOTE: this is distinct from handling in stage because it's not necessarily
                    # an error here if one of the previous records doesn't have a particular input,
                    # and we already warn about that in the previous conditional.
                    if function_input in prev_record.state:
                        # populate the dictionary associated with this input and record
                        # note that we keep resolve off, same as in stage, to keep it the
                        # lazy instance (since we don't know if this is actually needed
                        # yet or not)
                        prev_record.state.resolve = False
                        function_inputs[function_input][
                            prev_record
                        ] = prev_record.state[function_input]
                        prev_record.state.resolve = True
            function_inputs.update(kwargs)

            # see note in stage
            if cachers is not None:
                # instantiate cachers if not already
                for i in range(len(cachers)):
                    cacher = cachers[i]
                    if type(cacher) == type:
                        cachers[i] = cacher()
                    # set the active record on the cacher as well as provide a default name
                    # (the name of the output)
                    cachers[i].set_record(record)
                    cachers[
                        i
                    ].stage = name  # set current stage name, so get_path is correct in later stages (particularly for lazy)
                    if cachers[i].name is None and cachers[i].path_override is None:
                        if type(outputs[i]) == Lazy:
                            cachers[i].name = outputs[i].name
                        else:
                            cachers[i].name = outputs[i]
            record.stage_cachers = cachers

            # at this point we've grabbed all information we would need if we're
            # just mapping out the stages, so return at this point.
            # TODO: 3/8/2023 - not true, we're not getting outputs
            if record.manager.map_mode:
                record.manager.stage_active = False
                # in order to get a more detailed map that accurately shows input/output names,
                # we need to create pseudo-state-artifact-representations for the outputs. Since
                # we obviously can't add the actual artifacts (there are none without running!),
                # we just add the string key and a name, so record __repr__ has something to use
                _get_output_representations_for_map(record, outputs, cachers, records)
                record.stage_cachers = None
                return record

            # determine if we need to execute this stage and handle any
            # cached values.
            execute_stage = False  # false by default so the non-dag condition below can distinguish whether
            # the dag condition actually set execute stage to true or if just default
            pre_cache_time_start = time.perf_counter()  # time to load from cache
            record.manager.lock()

            # if we have an execution list from our stage DAG, use that
            # to determine if this stage executes or not.
            if record.manager.map is not None:
                stage_rep = (record.get_record_index(), name)
                if stage_rep not in record.manager.map.execution_list:
                    logging.debug('DAG-indicated stage skip "%s".' % str(stage_rep))
                    _dag_skip_check_cached_outputs(
                        name, record, outputs, cachers, records
                    )

                    # grab any possible previous reportables so they still end up in report.
                    _check_cached_reportables(name, record)
                    record.store_tracked_paths()
                    execute_stage = False
                else:
                    # the representation is in the execution list, so execute!
                    execute_stage = True

            # otherwise, proceed normally with cache and load checks
            # NOTE: this is an explicit _separate_ check because if our DAG indicates that
            # this stage executes, we still want to actually double check the cache and determine
            # if we _still_ need to execute this stage. This is primarily for cases where a
            # stage happens to get run multiple times with different records that have the same args/
            # same outputs. The DAG doesn't dynamically update with cached values during the actual
            # experiment run, so it won't catch this case by itself.
            # So the flow is: if we have a DAG, and it says to execute this stage, or if we don't have
            # a DAG, check if we actually need to run this based on cached values.
            if record.manager.map is None or execute_stage:
                cache_valid = _check_cached_outputs(
                    name, record, outputs, cachers, records
                )
                if cache_valid:
                    # get previous reportables if available
                    _check_cached_reportables(name, record, records)

                    # if we've hit this point, we will be returning early/not executing
                    # the stage because all outputs are found. The process of checking
                    # cached outputs should correctly add all the necessary tracked paths,
                    # so to transfer these paths into a store full run, we just need
                    # the below call
                    # NOTE: I _believe_ that we also get metadata from this because
                    # the load_metadata will have entered the metadata path already.
                    # this will need to be tested
                    record.store_tracked_paths()
                    execute_stage = False
                else:
                    # at least one output wasn't cached, so execute order 66!
                    execute_stage = True

            # check each input for Lazy objects and load them if we know we have to execute this stage
            if execute_stage:
                for function_input in function_inputs:
                    for prev_record, artifact in function_inputs[
                        function_input
                    ].items():
                        if isinstance(artifact, Lazy) and artifact.resolve:
                            logging.debug(
                                "Resolving lazy load object '%s' from record %s"
                                % (function_input, prev_record.get_reference_name())
                            )
                            function_inputs[function_input][
                                prev_record
                            ] = function_inputs[function_input][prev_record].load()

            record.manager.unlock()
            pre_cache_time_end = time.perf_counter()

            if not execute_stage:
                post_mem_usage = psutil.Process().memory_info().rss
                post_footprint = 0
                if os.name != "nt":
                    post_footprint = (
                        resource.getrusage(resource.RUSAGE_SELF).ru_maxrss * 1024
                    )
                _log_stats(
                    record,
                    pre_cache_time_start,
                    pre_cache_time_end,
                    pre_mem_usage,
                    post_mem_usage,
                    0,
                    0,
                    0,
                    0,
                    pre_footprint,
                    post_footprint,
                )
                utils.set_logging_prefix("")
                record.manager.stage_active = False
                record.manager.update_map_progress(record, "continue")
                record.stage_cachers = None
                return record

            # run the function
            logging.info("Stage (aggregate) %s executing...", name)
            exec_time_start = time.perf_counter()
            # NOTE: passing additional args to an aggregate stage is sort of undefined functionality,
            # this probably should be discouraged.
            function_outputs = function(record, records, **function_inputs)
            exec_time_end = time.perf_counter()

            # handle storing outputs in record
            post_cache_time_start = time.perf_counter()
            record.manager.lock()
            _store_outputs(name, record, outputs, cachers, function_outputs, records)
            _store_reportables(name, record, records)
            record.store_tracked_paths()
            record.manager.unlock()
            post_cache_time_end = time.perf_counter()

            # free up any memory from cached things
            cleaned_function_outputs = []
            lazy_found = False

            # iterate the outputs rather than function_outputs
            for index, output_name in enumerate(outputs):
                if type(output_name) == Lazy:
                    lazy_found = True
                    logging.debug("Lazy object '%s' will be cleaned." % output_name)
                    if index == 0 and len(outputs) == 1:
                        cleaned_function_outputs = outputs[
                            index
                        ]  # so that this doesn't syntactically read output_name, because it's Lazy not a str
                    else:
                        cleaned_function_outputs.append(outputs[index])
                else:
                    if index == 0 and len(outputs) == 1:
                        cleaned_function_outputs = function_outputs
                    else:
                        cleaned_function_outputs.append(function_outputs[index])
            if len(outputs) > 1:
                cleaned_function_outputs = tuple(cleaned_function_outputs)

            # free up any lazy cache objects
            if lazy_found:
                logging.debug("Freeing memory from lazy objects...")
                pre_del_mem_usage = psutil.Process().memory_info().rss
                del function_outputs
                post_del_mem_usage = psutil.Process().memory_info().rss
                del_mem_diff = pre_del_mem_usage - post_del_mem_usage
                logging.debug("Freed %s" % utils.human_readable_mem_usage(del_mem_diff))

            logging.info("Stage (aggregate) %s complete", name)

            # check memory usage
            post_footprint = 0
            if os.name != "nt":
                post_footprint = (
                    resource.getrusage(resource.RUSAGE_SELF).ru_maxrss * 1024
                )
            post_mem_usage = psutil.Process().memory_info().rss

            _log_stats(
                record,
                pre_cache_time_start,
                pre_cache_time_end,
                pre_mem_usage,
                post_mem_usage,
                exec_time_start,
                exec_time_end,
                post_cache_time_start,
                post_cache_time_end,
                pre_footprint,
                post_footprint,
            )

            record.output = cleaned_function_outputs
            utils.set_logging_prefix("")
            record.manager.stage_active = False
            record.manager.update_map_progress(record, "continue")
            record.stage_cachers = None
            return record

        return wrapper

    return decorator


def _get_output_representations_for_map(
    record, outputs: list[str], cachers: list[Cacheable], records: list[Record] = None
) -> list[MapArtifactRepresentation]:
    """Check if the requested outputs are cached but do not load them, simply return
    a list of statuses for found."""
    if cachers is not None and len(cachers) == 0:
        raise EmptyCachersError(
            "Do not use '[]' for cachers. This will always short-circuit because there is nothing that isn't cached."
        )

    artifacts = []
    for i in range(len(outputs)):
        metadata = None
        cacher = None
        is_cached = False
        if cachers is not None:
            cacher = cachers[i]
            is_cached = cacher.check()
            if is_cached:
                metadata = cacher.load_metadata()
        artifact = _add_output_artifact(
            record, None, outputs, i, metadata, cacher, is_cached
        )
        artifacts.append(artifact)

    return artifacts


def _dag_skip_check_cached_outputs(
    stage_name: str,
    record,
    outputs: list[str],
    cachers: list[Cacheable],
    records: list[Record] = None,
):
    """Checks for cached values and loads any relevant metadata into artifact representations.
    This function does not actually load values themselves, meant to be called for a skipped stage
    in dag-based execution."""
    if cachers is not None and len(cachers) == 0:
        raise EmptyCachersError(
            "Do not use '[]' for cachers. This will always short-circuit because there is nothing that isn't cached."
        )

    skipped_function_outputs = []
    for i, output_name in enumerate(outputs):
        cacher = None
        metadata = None
        is_cached = False
        output = None
        if cachers is not None:
            if cachers[i].check():
                cachers[i].load_metadata()
                if isinstance(outputs[i], Lazy):
                    outputs[i].cacher = cachers[i]
                    # we set the output to just be the Lazy instance for now
                    output = outputs[i]
                else:
                    # TODO: make a lazy for it anyway? Evnetually use ref instead
                    output = Lazy(output_name)
                    output.cacher = cachers[i]

                cacher = cachers[i]
                metadata = cachers[i].metadata
                is_cached = True
        artifact = _add_output_artifact(
            record, None, outputs, i, metadata, cacher, is_cached
        )
        if is_cached:
            artifact.file = cachers[i].get_path()
        # add a skippedoutput instance if it's not a cached value
        if output is None:
            output = SkippedOutput(record, stage_name, artifact)
        skipped_function_outputs.append(output)
        record.state[str(outputs[i])] = output


def _check_cached_outputs(
    stage_name: str,
    record: Record,
    outputs: list[Union[str, Lazy]],
    cachers: list[Cacheable],
    records: list[Record] = None,
) -> bool:
    """Run the ``.check()`` on each cacher to see if the artifacts are in the cache or not/
    if overwrite was specified and we need to re-run the stage.

    NOTE: this function _does_ load metadata into the cachers if found.
    """
    if cachers is not None and len(cachers) == 0:
        raise EmptyCachersError(
            "Do not use '[]' for cachers. This will always short-circuit because there is nothing that isn't cached."
        )

    cache_valid = False
    if cachers is not None:
        # NOTE: we put this here because the cachers still need to be set up
        if stage_name in record.manager.overwrite_stages or record.manager.overwrite:
            return False

        # check the cache (and load into record if found)
        cache_valid = True
        function_outputs = []
        for i in range(len(cachers)):
            if cachers[i].check():
                cachers[i].load_metadata()
                # handle lazy objects by setting the cacher but not actually loading yet.
                if type(outputs[i]) == Lazy:
                    outputs[i].cacher = cachers[i]
                    # we set the output to just be the Lazy instance for now
                    output = outputs[i]
                else:
                    output = cachers[i].load()

                function_outputs.append(output)
                record.state[str(outputs[i])] = output

                artifact = _add_output_artifact(
                    record,
                    output,
                    outputs,
                    i,
                    metadata=cachers[i].metadata,
                    cacher=cachers[i],
                    is_cached=True,
                )
                # TODO: (3/21/2023) possibly have "files" which would be cachers.cached_files?
                artifact.file = cachers[i].get_path()
            else:
                # we found something that wasn't cached, recompute everything
                cache_valid = False
                break

        # if we have the cached objects, return them right away
        if cache_valid:
            if len(cachers) == 1:
                record.output = function_outputs[0]
            else:
                record.output = tuple(function_outputs)

    return cache_valid


def _add_output_artifact(
    record,
    object: Any,
    outputs: list[Union[str, Lazy]],
    index: int,
    metadata=None,
    cacher=None,
    is_cached=False,
):
    """Manage representation recording - this creates an artifact representation and adds
    to the manager's artifacts."""
    if not record.manager.map_mode:
        artifact = ArtifactRepresentation(
            record, outputs[index], object, metadata=metadata, cacher=cacher
        )
    else:
        artifact = MapArtifactRepresentation(
            # NOTE: we don't use map=True for the get record index because we're still in map mode...
            # this should probably change at some point, where getindex also takes the current map_mode
            # into account
            record.get_record_index(),
            record.manager.current_stage_name,
            outputs[index],
            is_cached,
            metadata,
            cacher,
        )
    new_index = len(record.manager.artifacts)
    record.manager.artifacts.append(artifact)
    # if new_index not in record.stage_outputs[-1]:
    record.stage_outputs[-1].append(new_index)
    record.state_artifact_reps[str(outputs[index])] = new_index
    return artifact


def _check_cached_reportables(stage_name, record, aggregate_records=None) -> bool:
    """Look for any previously cached reportables"""
    reportables_list_cacher = FileReferenceCacher(
        name="reportables_file_list", record=record
    )
    if reportables_list_cacher.check():
        reportables_list_cacher.load_metadata()
        paths = reportables_list_cacher.load()
        for path in paths:
            with open(path, "rb") as infile:
                logging.debug("Reusing cached reportable '%s'" % path)
                reportable = pickle.load(infile)
                record.report(reportable)
        return True
    # the return of this function is simply used to determine if we need to store again?
    return False


def _store_reportables(stage_name, record, aggregate_records=None):
    """Save a copy of each reportable that comes from a stage, so that if the stage is re-run and
    doesn't actually execute (because outputs were cached), the reportables are still loaded and
    added to the report."""
    # get all reportables from the manager for this record and stage name
    reportables = []
    for reportable in record.manager.reportables:
        if reportable.record == record and reportable.stage == stage_name:
            reportables.append(reportable)

    if len(reportables) == 0:
        return

    # pickle each one and store it.
    paths = []
    reportables_path = record.get_dir(
        "reportables"
    )  # this will make sure all reportables go to full store
    for reportable in reportables:
        # make a copy of the reportable without the record, because that seems to break the mp.lock
        # when in parallel mode.
        # NOTE: do NOT use a deepcopy below, runs into same issue.
        reportable_copy = copy.copy(reportable)
        reportable_copy.record = None
        reportable_path = os.path.join(
            reportables_path, f"{reportable.qualified_name}.pkl"
        )
        paths.append(reportable_path)
        logging.debug("Caching reportable '%s'" % reportable_path)
        with open(reportable_path, "wb") as outfile:
            pickle.dump(reportable_copy, outfile)

    # write a cache file out containing the reportables path names.
    reportables_list_cacher = FileReferenceCacher(
        name="reportables_file_list", record=record
    )
    reportables_list_cacher.save(paths)

    # send along metadata for it, to track when the reportables were generated.
    # NOTE: we've already collected_metadata from passing record in init up above,
    # so we don't need to use the extra_metadata field on the cacher.
    reportables_list_cacher.metadata["extra"]["reportables"] = True
    reportables_list_cacher.save_metadata()


def _store_outputs(
    function_name,
    record: Record,
    outputs: list[Union[str, Cacheable]],
    cachers: list[Cacheable],
    function_outputs: list[any],
    records: list[Record] = None,
):
    """Store the stage outputs in the cache as appropriate."""
    if len(outputs) == 0:
        return

    if cachers is not None:
        logging.info("Stage %s caching outputs..." % function_name)

    if type(function_outputs) != tuple:
        function_outputs = (function_outputs,)

    if len(outputs) != len(function_outputs):
        raise OutputSignatureError(
            "Returned values from '%s' do not match expected stage outputs. The function should return values for %s"
            % (function_name, str(outputs))
        )

    # store each argument in the record and cache if requested
    for index, output in enumerate(function_outputs):
        if type(outputs[index]) == Lazy:
            record.state[str(outputs[index])] = outputs[index]
            # TODO: (01/13/2022) if cachers is none, throw an error
        else:
            record.state[str(outputs[index])] = output

        # manage representation recording
        artifact = _add_output_artifact(record, output, outputs, index)
        if (
            cachers is not None
            and not record.manager.dry
            and not record.manager.dry_cache
        ):
            logging.debug(
                f"Caching {outputs[index]} to '{cachers[index].get_path()}'..."
            )
            cachers[index].save(output)
            artifact.file = cachers[index].get_path()

            # generate and save metadata
            # note that if we got to this point, we actually ran the stage code, so
            # we generate _new_ metadata
            cachers[index].collect_metadata()
            cachers[index].metadata["preview"] = artifact.string
            metadata = cachers[index].save_metadata()
            artifact.metadata = metadata

        # if specified as lazy, be sure to populate the cacher
        if type(outputs[index]) == Lazy:
            outputs[index].cacher = cachers[index]



"""This file is simply a manual way into the experiment script
from curifactory - this allows running it from inside an IDE
such as pycharm, in order to get debugging."""

from curifactory import experiment

experiment.main()



"""Contains the base parameter class ExperimentParameters, a dataclass meant to represent
a configuration for an experiment run."""

from dataclasses import dataclass, field
from typing import Callable, Union

from curifactory import hashing


@dataclass
class ExperimentParameters:
    """Base parameter class, for handling naming and hashing.

    In any given repo, this class should be extended to contain any needed
    local configuration.

    .. note::

        Extending with a ``@dataclass`` is recommended to make it syntactically
        easier to read and define.

    Example:
        .. code-block:: python

            from dataclasses import dataclass
            from curifactory import ExperimentParameters

            @dataclass
            class Params(ExperimentParameters):
                some_parameter: int = 0
                # ...
    """

    name: str = "UNNAMED"
    """Parameter set name. This can be used to easily distinguish/refer to specific
        configurations in aggregate stages. This should be unique for every parameter set."""
    hash: str = None
    """Curifactory automatically fills this, but it can be overriden if you need to use
        very specific cache naming. (Should not normally be necessary.)"""
    overwrite: bool = False
    """Whether to overwrite pre-cached values. Curifactory automatically sets this based
        on command line flags."""

    # NOTE: may be able to take parent classes' hashing functions into account as well
    # https://stackoverflow.com/questions/10091957/get-parent-class-name
    hash_representations: dict[str, Union[None, Callable]] = field(
        default_factory=dict, repr=False
    )
    """Dictionary of parameter names in the dataclass where you can provide functions
        that return a unique/consistent representation for each parameter to use as
        the value for hashing.  Setting these allow control over what counts as a
        cache miss/cache hit. Sane defaults are applied to any parameters that don't
        have a representation function specified here. (see ``hashing.get_parameter_hash_value``)

        Any function provided is expected to take self (the entire parameter set instance) and
        the value of the named parameter to be hashed.

        You can also provide ``None`` for any parameters in order to exclude their value
        from the hash entirely. (This is useful for operational arguments e.g. the number
        of GPU's or processes to run on, where you would not want changing that to invalidate
        the existing cached values.)
    """

    # TODO: implement __eq__ based on args_hash

    def params_hash(self, dry=False):
        """Convenience function to see the hash of these parameters that curifactory is
        computing, or debug them with ``dry=True``."""
        return hashing.hash_param_set(self, store_in_registry=False, dry=dry)
